/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2014-2019, 2021, 2023 Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

load "s7-1200-basic".
load "lib-conv".

set outfile="/dev/null".

set print_states=false.    
       
/* the following ladder logic 
        In0.0         In|In|.8      Out0.0
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0.0         In|In|.8      Out0.1    |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0.0         In|In|.8      Out0.2    |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0.0         In|In|.8      Out0.3    |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    .
    .
    .
    |   In0.0         In|In|.8      Out|Out|.8|
    
    is translated into Tempura using the constructs defined in s7-1200-basic.t
    
 */
define plc(In,Out) = {
  /* let's output the value of In and Out in each state */
  /*always ( format("In=%t, Out=%t\n",In,Out) )  and*/
  for i<|Out| do {
   for j<8 do {               
     /* rung 8*i+j */
     COIL(AND_IL(all(In)),Out[i][j]) and skip and no_ch_list(In) and
     forall k<|Out| : {
        if k=i then {
          forall l<8 : { if l ~= j then stable(Out[k][l]) }
        }
        else {
          stable(Out[k])
        }
     }
  }
 }
}.  
    

/* X : used to store date read from external file
   I: input I.0 .. I.ni to the plc each Ii is 8 bits
   Q: output Q.0..Q.nq from the plc each Qi is 8 bits
   PL: expected output PL.0 .. PL.nq from formal spec each PLi is 8 bits
*/   
define test_plc(ni,nq) = { 
   exists X,I,Q,PL : {
    {set infile="S71200-test.TXT" and empty};
    {   
    /* read the first data element from the file */	   
    input X  and
    /* ni inputs and nq outputs */               
    list(I,ni) and stable(struct(I)) and
    list(Q,nq) and stable(struct(Q)) and              
    list(PL,nq) and stable(struct(PL)) and
    /* PLi is 8 bits, Ii is 8 bits, Qi is 8 bits */ 
    (forall i<nq: {
            list(Q[i],8) and stable(struct(Q[i])) and
            list(PL[i],8) and stable(struct(PL[i])) and    
    /* Qi and PLi are initially all zeros */
     Q[i]=[0,0,0,0,0,0,0,0] and
    PL[i]=[0,0,0,0,0,0,0,0]
    } ) and
    (forall i<ni: {
            list(I[i],8) and stable(struct(I[i])) and   
    /* Ii is initially all zeros */
    I[i]= [0,0,0,0,0,0,0,0] 
    } ) and
    /* as long the input is not zero we continue otherwise we stop */
    while (X ~= 0) do {
       /* we discard the first input as it is MK: data */
       {skip and no_ch_list(I) and no_ch_list(Q) and no_ch_list(PL)};
       /* the second data element is input to the plc and stored in I */
       { input X  and skip and
         if X[0..3]="DI:" then {
           (forall i<ni:{  I[i]:=inttobits(hexbinbyte(X[3+i*3],X[4+3*i]))})
         } and
         no_ch_list(Q) and no_ch_list(PL)
       };
       /* the third data element is output from the plc and stored in Q */
       { input X  and skip and
         if X[0..3]="DO:" then {
           (forall i<nq:{  Q[i]:=inttobits(hexbinbyte(X[3+i*3],X[4+3*i]))})
         } and
         no_ch_list(I) and no_ch_list(PL)
       }; /* the fourth data element is CT data and discarded */
       { input X and skip and
         no_ch_list(I) and no_ch_list(Q) and no_ch_list(PL)
       }; /* the fifth data element is TI data and discarded */
       { input X  and skip and
         no_ch_list(I) and no_ch_list(Q) and no_ch_list(PL)
       }; /* we feed the input I to our formal specified plc and store
             the output in PL */
       { plc(I,PL) and no_ch_list(I) and no_ch_list(Q) 
       }; /* we start all over again but also check whether the
             received output Q matches the output PL from the
	     formal specification */
       { input X  and empty and
         forall j<nq : {
             forall i<8 : {
          if PL[j][i] ~= Q[j][i] then {
                  format("Output Q[%d][%d]=%t does not match expected output PL[%d][%d]=%t\n",j,i,Q[j][i],j,i,PL[j][i])
	  }
	 }
        }   
       }	  
    }
  }
 }   
}.


message("Testing",1,1).
set outfile="/dev/null".
run test_plc(1, 1).
message("Testing",1,2).
set outfile="/dev/null".    
run test_plc(1, 2).
message("Testing",1,4).
set outfile="/dev/null".    
run test_plc(1, 4).    
message("Testing",1,8).
set outfile="/dev/null".
run test_plc(1, 8).
message("Testing",1,16).
set outfile="/dev/null".
run test_plc(1,16).

message("Testing",2,1).
set outfile="/dev/null".
run test_plc(2, 1).
message("Testing",2,2).
set outfile="/dev/null".    
run test_plc(2, 2).
message("Testing",2,4).
set outfile="/dev/null".    
run test_plc(2, 4).    
message("Testing",2,8).
set outfile="/dev/null".
run test_plc(2, 8).
message("Testing",2,16).
set outfile="/dev/null".
run test_plc(2,16).    

message("Testing",4,1).
set outfile="/dev/null".
run test_plc(4, 1).
message("Testing",4,2).
set outfile="/dev/null".    
run test_plc(4, 2).
message("Testing",4,4).
set outfile="/dev/null".    
run test_plc(4, 4).    
message("Testing",4,8).
set outfile="/dev/null".
run test_plc(4, 8).
message("Testing",4,16).
set outfile="/dev/null".
run test_plc(4,16).    

message("Testing",8,1).
set outfile="/dev/null".
run test_plc(8, 1).
message("Testing",8,2).
set outfile="/dev/null".    
run test_plc(8, 2).
message("Testing",8,4).
set outfile="/dev/null".    
run test_plc(8, 4).    
message("Testing",8,8).
set outfile="/dev/null".
run test_plc(8, 8).
message("Testing",8,16).
set outfile="/dev/null".
run test_plc(8,16).

message("Testing",16,1).
set outfile="/dev/null".
run test_plc(16, 1).
message("Testing",16,2).
set outfile="/dev/null".    
run test_plc(16, 2).
message("Testing",16,4).
set outfile="/dev/null".    
run test_plc(16, 4).    
message("Testing",16,8).
set outfile="/dev/null".
run test_plc(16, 8).
message("Testing",16,16).
set outfile="/dev/null".
run test_plc(16,16).
    
set outfile="stdout".    
