/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: init.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"




/*******************\
*                   *
* GLOBAL VARIABLES  *
*                   *
\*******************/

/* Static values (very commonly used) */
struct node TrueNode = {T,BOOL,NULL,{.bval=T}};
NODEPTR True = &TrueNode;
struct node FalseNode = {T,BOOL,NULL,{.bval=F}};
NODEPTR False = &FalseNode;
struct node ZeroNode = {T,INT,NULL,{.ival=0}};
NODEPTR Zero = &ZeroNode;
struct node OneNode = {T,INT,NULL,{.ival=1}};
NODEPTR One = &OneNode;
struct node NilNode = {T,NIL,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Nil = &NilNode;
struct node EmptyNode = {T,EMPTY,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Empty = &EmptyNode;
struct node MoreNode = {T,MORE,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR More = &MoreNode;
struct node SkipNode = {T,SKIP,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Skip = &SkipNode;
struct node RanlenNode = {T,RANLEN,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Ranlen = &RanlenNode;
struct node InfiniteNode = {T,INFINITE,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Infinite = &InfiniteNode;
struct node StateNrNode = {T,STATENUMBER,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR StateNr = &StateNrNode;
struct node RandomNode = {T,RANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Random = &RandomNode;
struct node RRandomNode = {T,RRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR RRandom = &RRandomNode;
struct node SRandomNode = {T,SRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR SRandom = &SRandomNode;
struct node SRRandomNode = {T,SRRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR SRRandom = &SRRandomNode;

struct node ZRandomNode = {T,ZRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR ZRandom = &ZRandomNode;
struct node ZRRandomNode = {T,ZRRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR ZRRandom = &ZRRandomNode;
struct node FRandomNode = {T,FRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR FRandom = &FRandomNode;
struct node FRRandomNode = {T,FRRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR FRRandom = &FRRandomNode;

struct node EndOfFileNode = {T,END,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR EndOfFile = &EndOfFileNode;
/*MEMORY GlobalEmpty[MEMCELLSIZE] = {{T,0,T,F,0,nodep (&TrueNode)},*/
/* 			           {T,1,T,F,0,nodep (&TrueNode)}};*/
NODE GlobalEmpty[MEMCELLSIZE];
NODEPTR EmptyLoc = GlobalEmpty;

unsigned Ready = F;                  /* True after successful reduction. */
unsigned Retry = F;                  /* Re-reduce in current pass? */
unsigned NextFlag = F;               /* Reduce next value of expression? */
unsigned int PrevFlag = 0; 
unsigned RedefFlag = F;              /* Redefinition allowed? */
unsigned Clock = 0;
unsigned AlarmGone = F;
unsigned IsInitial = T;
/* counter for top-level commands */
unsigned Counter = 1;
unsigned INTERACTIVE;
unsigned EOFILE;
unsigned INSCANSTR;



/* Memory */
NODE GlobalMemory[offset2 (GLOBALMEMSIZE,1)];
unsigned LocalMemUsed = 0;
/*MEMORY TempMem[MEMCELLSIZE] = {{T,0,T,F,0,NULL},*/
/* 			       {T,1,T,F,0,NULL}};*/
NODE TempMem[MEMCELLSIZE];
NODEPTR TempLoc = TempMem;
NODEPTR iRememberList = NULL; /* initially we have no iREMEMBER statements */

unsigned CurrBit = 0;
unsigned NextBit = 1;

unsigned NoFraming = 0;
unsigned Framing = 1;


unsigned OneLevelMemFlag = F;        /* if on, use one-level memory */
unsigned TimeStampCheckFlag = T;     /* if on, check time stamp */

/* environment */
NODE GlobalEnv = {T,ENV};
 
                         
ENVSTACK EnvStack[MAXENVSTACKSIZE] = {{GlobalMemory, &GlobalEnv}};
int EnvTop = 0;
int EnvBase = 0;

/* reduce-stack */
REDUCESTACK ReduceStack[MAXREDUCESTACKSIZE];
REDUCESTACKPTR ReduceStackTop = ReduceStack+(-1);
REDUCESTACKPTR ReduceStackMax = ReduceStack+(MAXREDUCESTACKSIZE-1);
NODEPTR LastCommand = &TrueNode;

/* tracing and debugging */
jmp_buf *BreakPoint;
int TempuraLevel = 0;
int BreakDepth = -1;
unsigned Break = F;
unsigned InterruptSet = F;
int PassLimitCounter = 0;
unsigned PassProgressFlag = F;
unsigned StateProgressFlag = T;
unsigned ForceOutput = F;
unsigned RunFlag = F;
/*int SyntaxErrorCount = 0;*/

/* statistics */
unsigned StateNumber = 0;
unsigned PassNumber = 0;
unsigned TotalPassCount = 0;
unsigned TotalReduceCount = 0;
unsigned TotalReduceTimes = 0;
unsigned StateReduceCount = 0;

unsigned PassReduceCount = 0;
unsigned TotalSuccessfulReduceCount = 0;
unsigned TotalSuccessfulReduceTimes = 0;
unsigned StateSuccessfulReduceCount = 0;
unsigned PassSuccessfulReduceCount = 0;
int ReduceDepth = 0;
int MaxReduceDepth = 0;

/* Heap */
HEAP Heap = NULL;
unsigned FreeNodes = 0;
unsigned TempNodes = 0;
unsigned HeapSize = 0;
NODEPTR FreeNodeList = NULL;
NODEPTR TempNodeList = NULL;

/* Symbol Table */
NODEPTR SymbolTableTop = 0;
NODEPTR SymbolTableMax = 0;
NODEPTR HashTable[HASHTABLESIZE];/* The hash table. */

/* Wait list for system calls */
NODEPTR SysWaitList = NULL;

/* System variables (user definable). */
int PassLimit = -1;
int Verbosity = 1;
int PrintLevel = 2; 
int DebugLevel = 1;
int64_t Precision = 7;
int FLRnd = 0;
mpfr_prec_t MPFRPrecision = 53;
mpfr_rnd_t  MPFRRnd = MPFR_RNDN;
int TMPFRRnd = 0;
unsigned StatePrintFlag = T;
unsigned RedundancyFlag = F;
unsigned CallByNameFlag = F;
unsigned AssignAheadFlag = T;
unsigned BreakIsAbortFlag = T;
unsigned HopChop = T;

NODEPTR Path;       /* BCM 14Dec96 */
NODEPTR HelpDir;    /* BCM 14Dec96 */
NODEPTR InFile;
NODEPTR OldInFile;
NODEPTR OutFile;
FILE *Fpcur = NULL;
char *InFilename = "Standard Input";


/* Previous two lines were as follows.  Had problems with garbage collection.
                BCM 14Dec96
char *Path;
char *HelpDir = "Help";
*/
int TimeStep = 0;
/* values related to random number generator */

int64_t MaxRandLen = 10;   /* Maximum random interval length. */

int64_t RandSeed = 0;      /* Seed for random numbers. */
uint64_t TRandSeed = 0;
int64_t StateNrVal = -1;
unsigned StateNrValTime = 0;
int64_t RandomVal = -1;     /* random value >=0 for current state.
 			   Initially "undefined" as -1. */
unsigned RandomValTime = 0; /* Time of random value. */

/*int64_t RRandSeed = 0;   
  int TRRandSeed = 0; */
int64_t RRandomVal = -1;     /* Random value >=0.
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned RRandomValTime = 0; /* Time of Random value. AC 20/09/2000 */

/*int64_t SRandSeed = 0;
  int TSRandSeed = 0; */
double SRandomVal = -1;     /* frandom value in [0.0,1.0).
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned SRandomValTime = 0; /* Time of frandom value. AC 20/09/2000 */

/*int64_t SRRandSeed = 0;
  int TSRRandSeed = 0;*/
double SRRandomVal = -1;     /* fRandom value in [0.0,1.0).
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned SRRandomValTime = 0; /* Time of fRandom value. AC 20/09/2000 */

gmp_randstate_t  Randstate;
mpz_t ZRandSeed;
mpz_t ZRandomMax;
mpz_t ZRandomVal;
mpz_t ZRRandomVal;
unsigned ZRandomValTime = 0;
unsigned ZRRandomValTime = 0;
mpfr_t FRandomVal;
mpfr_t FRRandomVal;
unsigned FRandomValTime = 0;
unsigned FRRandomValTime = 0;





/* Input/Output */
/*NODEPTR AuxVal;*/
NODEPTR Parseroot = NULL;
/*int yydebug;*/
int yyerrorcount=0;
YYLTYPE yylloc;

int CharsOutput = 0;
FILE *InputFile; /*={stdin};*/
/*char InputText[MAXTOKENSIZE];*/
char *InputFilename = "Standard Input";

char *Keyword[MAXTOKEN - MINTOKEN];
PFI Reduces[MAXTOKEN - MINTOKEN][3];
char *Reduc[MAXTOKEN - MINTOKEN][3];

/* deleted 

 */
/*unsigned Token = END;*/
/*unsigned Recover = F;*/

/* #define MAXREDUCETOKEN COMMA-MINTOKEN */

/************************************************************\
* 							     *
* This routine enters reserved words into the symbol table   *
* and initialises reduction table                            *
* 							     *
\************************************************************/
    
void initialise_symboltable(void)
{
        /*  */    
     install_keyword(END,".");
     install_reduction(END, reduce_EOF,reduce_EOF,reduce_EOF );
     install_reduction_w(END,"reduce_EOF","reduce_EOF","reduce_EOF");
         /*  */
     install_keyword(NIL,"nil");
     install_reduction(NIL, assert_expr,reduce_const,get_ref );
     install_reduction_w(NIL,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(BOOL,"bool");
     install_reduction(BOOL, assert_true,reduce_const,get_ref );
     install_reduction_w(BOOL,"assert_true","reduce_const","get_ref");
         /*  */
     install_keyword(INT,"integer");
     install_reduction(INT, assert_expr,reduce_const,get_ref );
     install_reduction_w(INT,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(LAMBDA,"lambda");
     install_reduction(LAMBDA, assert_expr,reduce_const,get_ref );
     install_reduction_w(LAMBDA,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(iLAMBDA,"_lambda");
     install_reduction(iLAMBDA, assert_expr,reduce_const,get_ref );
     install_reduction_w(iLAMBDA,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(STRING,"string");
     install_reduction(STRING, assert_expr,reduce_const,get_ref );
     install_reduction_w(STRING,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(FLOAT,"float");
     install_reduction(FLOAT, assert_expr,reduce_const,get_ref );
     install_reduction_w(FLOAT,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(LIST,"list");
     install_reduction(LIST, assert_expr,reduce_const,get_ref );
     install_reduction_w(LIST,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(STATE,"state");
     install_reduction(STATE, assert_true,reduce_id_val,reduce_id_ref );
     install_reduction_w(STATE,"assert_true","reduce_id_val","reduce_id_ref");
         /*  */
     install_keyword(STATIC,"static");
     install_reduction(STATIC, assert_true,reduce_id_val,reduce_id_ref );
     install_reduction_w(STATIC,"assert_true","reduce_id_val","reduce_id_ref");
         /*  */
     install_keyword(EMPTY,"empty");
     install_reduction(EMPTY, reduce_EMPTY_stmt,reduce_EMPTY_val,reduce_EMPTY_ref );
     install_reduction_w(EMPTY,"reduce_EMPTY_stmt","reduce_EMPTY_val","reduce_EMPTY_ref");
         /*  */
     install_keyword(SKIP,"skip");
     install_reduction(SKIP, reduce_SKIP, assert_stmt, assert_stmt );
     install_reduction_w(SKIP,"reduce_SKIP","assert_stmt","assert_stmt");
         /*  */
     install_keyword(AND,"and");
     install_reduction(AND, reduce_AND,reduce_bin_op,get_ref );
     install_reduction_w(AND,"reduce_AND","reduce_bin_op","get_ref");
         /*  */
     install_keyword(iAND,"_and");
     install_reduction(iAND, reduce_iAND,assert_stmt,assert_stmt );
     install_reduction_w(iAND,"reduce_iAND","assert_stmt","assert_stmt");
         /*  */
     install_keyword(OR,"or");
     install_reduction(OR, reduce_OR,reduce_bin_op,get_ref );
     install_reduction_w(OR,"reduce_OR","reduce_bin_op","get_ref");
         /*  */
     install_keyword(NOT,"not");
     install_reduction(NOT, assert_true,reduce_u_op,get_ref );
     install_reduction_w(NOT,"assert_true","reduce_u_op","get_ref");
         /*  */
     install_keyword(IF,"if");
     install_reduction(IF, reduce_IF_stmt,reduce_IF_val,get_ref );
     install_reduction_w(IF,"reduce_IF_stmt","reduce_IF_val","get_ref");
         /*  */
     install_keyword(CALL,"_call");
     install_reduction(CALL, reduce_CALL,reduce_CALL,reduce_CALL );
     install_reduction_w(CALL,"reduce_CALL","reduce_CALL","reduce_CALL");
         /*  */
     install_keyword(iCALL,"_icall");
     install_reduction(iCALL, reduce_iCALL_stmt,reduce_iCALL_val,reduce_iCALL_ref );
     install_reduction_w(iCALL,"reduce_iCALL_stmt","reduce_iCALL_val","reduce_iCALL_ref");
         /*  */
     install_keyword(EXISTS,"exists");
     install_reduction(EXISTS, reduce_EXISTS,reduce_EXISTS,reduce_EXISTS );
     install_reduction_w(EXISTS,"reduce_EXISTS","reduce_EXISTS","reduce_EXISTS");
         /*  */
     install_keyword(iEXISTS,"_exists");
     install_reduction(iEXISTS, reduce_iEXISTS_stmt,reduce_iEXISTS_val,get_ref );
     install_reduction_w(iEXISTS,"reduce_iEXISTS_stmt","reduce_iEXISTS_val","get_ref");
         /*  */
     install_keyword(PROCESS,"process");
     install_reduction(PROCESS, reduce_PROCESS,assert_stmt,assert_stmt);
     install_reduction_w(PROCESS,"reduce_PROCESS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iPROCESS,"_process");
     install_reduction(iPROCESS, reduce_iPROCESS,assert_stmt,assert_stmt );
     install_reduction_w(iPROCESS,"reduce_iPROCESS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(PREFIX,"prefix");
     install_reduction(PREFIX, reduce_PREFIX,assert_stmt,assert_stmt );
     install_reduction_w(PREFIX,"reduce_PREFIX","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iPREFIX,"_prefix");
     install_reduction(iPREFIX, reduce_iPREFIX,assert_stmt,assert_stmt );
     install_reduction_w(iPREFIX,"reduce_iPREFIX","assert_stmt","assert_stmt");
         /*  */
     install_keyword(FORALL,"forall");
     install_reduction(FORALL, reduce_FORALL,reduce_FORALL,reduce_FORALL );
     install_reduction_w(FORALL,"reduce_FORALL","reduce_FORALL","reduce_FORALL");
         /*  */
     install_keyword(THE,"the");
     install_reduction(THE, assert_expr,reduce_THE,reduce_THE );
     install_reduction_w(THE,"assert_expr","reduce_THE","reduce_THE");
         /*  */
     install_keyword(ALWAYS,"always");
     install_reduction(ALWAYS, reduce_ALWAYS,assert_stmt,assert_stmt );
     install_reduction_w(ALWAYS,"reduce_ALWAYS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SOMETIME,"sometimes");
     install_reduction(SOMETIME, reduce_SOMETIME,assert_stmt,assert_stmt );
     install_reduction_w(SOMETIME,"reduce_SOMETIME","assert_stmt","assert_stmt");
         /*  */
     install_keyword(HALT,"halt");
     install_reduction(HALT, reduce_HALT,assert_stmt,assert_stmt );
     install_reduction_w(HALT,"reduce_HALT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(INIT,"init");
     install_reduction(INIT,reduce_INIT,assert_stmt,assert_stmt  );
     install_reduction_w(INIT,"reduce_INIT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(FIN,"fin");
     install_reduction(FIN, reduce_FIN,assert_stmt,assert_stmt );
     install_reduction_w(FIN,"reduce_FIN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(KEEP,"keep");
     install_reduction(KEEP, reduce_KEEP,assert_stmt,assert_stmt );
     install_reduction_w(KEEP,"reduce_KEEP","assert_stmt","assert_stmt");
         /*  */
     install_keyword(NEXT,"next");
     install_reduction(NEXT, reduce_NEXT_stmt,reduce_NEXT_val,reduce_NEXT_ref );
     install_reduction_w(NEXT,"reduce_NEXT_stmt","reduce_NEXT_val","reduce_NEXT_ref");
         /*  */
     install_keyword(WNEXT,"wnext");
     install_reduction(WNEXT, reduce_WNEXT,assert_stmt,assert_stmt );
     install_reduction_w(WNEXT,"reduce_WNEXT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(LEN,"len");
     install_reduction(LEN, reduce_LEN,assert_stmt,assert_stmt );
     install_reduction_w(LEN,"reduce_LEN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iLEN,"_len");
     install_reduction(iLEN, reduce_iLEN,assert_stmt,assert_stmt );
     install_reduction_w(iLEN,"reduce_iLEN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(LT,"<");
     install_reduction(LT, assert_stmt_to_true,reduce_bin_op,get_ref );
     install_reduction_w(LT,"assert_stmt_to_true","reduce_bin_op","get_ref");
         /*  */
     install_keyword(LE,"<=");
     install_reduction(LE, assert_stmt_to_true,reduce_bin_op,get_ref );
     install_reduction_w(LE,"assert_stmt_to_true","reduce_bin_op","get_ref");
         /*  */
     install_keyword(EQ,"=");
     install_reduction(EQ, reduce_EQ_stmt,reduce_EQ_val,get_ref );
     install_reduction_w(EQ,"reduce_EQ_stmt","reduce_EQ_val","get_ref");
         /*  */
     install_keyword(GE,">=");
     install_reduction(GE, assert_stmt_to_true,reduce_bin_op,get_ref );
     install_reduction_w(GE,"assert_stmt_to_true","reduce_bin_op","get_ref");
         /*  */
     install_keyword(GT,">");
     install_reduction(GT, assert_stmt_to_true,reduce_bin_op,get_ref );
     install_reduction_w(GT,"assert_stmt_to_true","reduce_bin_op","get_ref");
         /*  */
     install_keyword(NEQ,"~=");
     install_reduction(NEQ, assert_stmt_to_true,reduce_EQ_val,get_ref );
     install_reduction_w(NEQ,"reduce_EQ_stmt","reduce_EQ_val","get_ref");
         /*  */
     install_keyword(GETS,"gets");
     install_reduction(GETS, reduce_GETS,assert_stmt,assert_stmt );
     install_reduction_w(GETS,"reduce_GETS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(STB,"stable");
     install_reduction(STB, reduce_STB,assert_stmt,assert_stmt );
     install_reduction_w(STB,"reduce_STB","assert_stmt","assert_stmt");
         /*  */
     install_keyword(WHERE,"where");
     install_reduction(WHERE, reduce_WHERE,assert_stmt,assert_stmt );
     install_reduction_w(WHERE,"reduce_WHERE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(IS,"is");
     install_reduction(IS, reduce_IS,assert_stmt,assert_stmt );
     install_reduction_w(IS,"reduce_IS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(ASSN,"<-");
     install_reduction(ASSN, reduce_ASSN,assert_stmt,assert_stmt );
     install_reduction_w(ASSN,"reduce_ASSN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(UASSN,":=");
     install_reduction(UASSN, reduce_UASSN,assert_stmt,assert_stmt );
     install_reduction_w(UASSN,"reduce_UASSN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SKIPUASSN,":-");
     install_reduction(SKIPUASSN, reduce_SKIPUASSN,assert_stmt,assert_stmt );
     install_reduction_w(SKIPUASSN,"reduce_SKIPUASSN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SEMI,";");
     install_reduction(SEMI, reduce_SEMI,assert_stmt,assert_stmt );
     install_reduction_w(SEMI,"reduce_SEMI","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iSEMI,"_;");
     install_reduction(iSEMI, reduce_iSEMI,assert_stmt,assert_stmt );
     install_reduction_w(iSEMI,"reduce_iSEMI","assert_stmt","assert_stmt");
         /*  */
     install_keyword(WHILE,"while");
     install_reduction(WHILE, reduce_WHILE,assert_stmt,assert_stmt );
     install_reduction_w(WHILE,"reduce_WHILE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(REPEAT,"repeat");
     install_reduction(REPEAT, reduce_REPEAT,assert_stmt,assert_stmt );
     install_reduction_w(REPEAT,"reduce_REPEAT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(FOR,"for");
     install_reduction(FOR, reduce_FOR,assert_stmt,assert_stmt );
     install_reduction_w(FOR,"reduce_FOR","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iFOR,"_for");
     install_reduction(iFOR, assert_expr,reduce_iFOR,assert_stmt );
     install_reduction_w(iFOR,"assert_expr","reduce_iFOR","assert_stmt");
         /*  */
     install_keyword(FORTIMES,"_for");
     install_reduction(FORTIMES, reduce_FORTIMES,assert_stmt,assert_stmt );
     install_reduction_w(FORTIMES,"reduce_FORTIMES","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iFORTIMES,"_for_times");
     install_reduction(iFORTIMES, assert_expr,reduce_iFORTIMES,assert_stmt );
     install_reduction_w(iFORTIMES,"assert_expr","reduce_iFORTIMES","assert_stmt");
         /*  */
     install_keyword(FORIN,"_for");
     install_reduction(FORIN, reduce_FORIN, assert_stmt, assert_stmt );
     install_reduction_w(FORIN,"reduce_FORIN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(FORALLIN,"_forall");
     install_reduction(FORALLIN, reduce_FORALLIN, reduce_FORALLIN, reduce_FORALLIN );
     install_reduction_w(FORALLIN,"reduce_FORALLIN","reduce_FORALLIN","reduce_FORALLIN");
         /*  */
     install_keyword(iFORIN,"_for-in");
     install_reduction(iFORIN, assert_expr,reduce_iFORIN,assert_stmt );
     install_reduction_w(iFORIN,"assert_expr","reduce_iFORIN","assert_stmt");
         /*  */
     install_keyword(PROJ,"proj");
     install_reduction(PROJ, reduce_PROJ,assert_stmt,assert_stmt );
     install_reduction_w(PROJ,"reduce_PROJ","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iPROJ,"_proj");
     install_reduction(iPROJ, reduce_iPROJ,assert_stmt,assert_stmt );
     install_reduction_w(iPROJ,"reduce_iPROJ","assert_stmt","assert_stmt");
         /*  */
     install_keyword(INPUT,"input");
     install_reduction(INPUT, reduce_INPUT,assert_stmt,assert_stmt );
     install_reduction_w(INPUT,"reduce_INPUT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(OUTPUT,"output");
     install_reduction(OUTPUT, reduce_OUTPUT,reduce_OUTPUT,assert_stmt );
     install_reduction_w(OUTPUT,"reduce_OUTPUT","reduce_OUTPUT","assert_stmt");
         /*  */
     install_keyword(PLUS,"+");
     install_reduction(PLUS, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(PLUS,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(MINUS,"-");
     install_reduction(MINUS, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MINUS,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(MULT,"*");
     install_reduction(MULT, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MULT,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(DIV,"div");
     install_reduction(DIV, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(DIV,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(MOD,"mod");
     install_reduction(MOD, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MOD,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(POWER,"**");
     install_reduction(POWER, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(POWER,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(UPLUS,"+");
     install_reduction(UPLUS, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(UPLUS,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(UMINUS,"-");
     install_reduction(UMINUS, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(UMINUS,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(MAKELIST,"_make_list");
     install_reduction(MAKELIST, assert_expr,reduce_MAKELIST,get_ref );
     install_reduction_w(MAKELIST,"assert_expr","reduce_MAKELIST","get_ref");
         /*  */
     install_keyword(MAKERANGELIST,"_make_range_list");
     install_reduction(MAKERANGELIST, assert_expr,reduce_MAKERANGELIST,get_ref );
     install_reduction_w(MAKERANGELIST,"assert_expr","reduce_MAKERANGELIST","get_ref");
         /*  */
     install_keyword(TYPE,"type");
     install_reduction(TYPE, assert_expr,reduce_TYPE,get_ref );
     install_reduction_w(TYPE,"assert_expr","reduce_TYPE","get_ref");
         /*  */
     install_keyword(STRUCT,"struct");
     install_reduction(STRUCT, assert_expr,reduce_STRUCT_val,reduce_STRUCT_ref );
     install_reduction_w(STRUCT,"assert_expr","reduce_STRUCT_val","reduce_STRUCT_ref");
         /*  */
     install_keyword(SIZE,"_size");
     install_reduction(SIZE, assert_expr,get_val,reduce_SIZE );
     install_reduction_w(SIZE,"assert_expr","get_val","reduce_SIZE");
         /*  */
     install_keyword(SUB,"_sub");
     install_reduction(SUB, assert_true,get_val,reduce_SUB );
     install_reduction_w(SUB,"assert_true","get_val","reduce_SUB");
         /*  */
     install_keyword(SUBLIST,"_sub_list");
     install_reduction(SUBLIST, assert_expr,get_val,reduce_SUBLIST );
     install_reduction_w(SUBLIST,"assert_expr","get_val","reduce_SUBLIST");
         /*  */
     install_keyword(FIXLIST,"list");
     install_reduction(FIXLIST, reduce_list,assert_stmt,assert_stmt );
     install_reduction_w(FIXLIST,"reduce_list","assert_stmt","assert_stmt");
         /*  */
     install_keyword(VARLIST,"List");
     install_reduction(VARLIST, reduce_list,assert_stmt,assert_stmt );
     install_reduction_w(VARLIST,"reduce_list","assert_stmt","assert_stmt");
         /*  */
     install_keyword(REFER,"ref");
     install_reduction(REFER, assert_expr,reduce_REFER,get_ref );
     install_reduction_w(REFER,"assert_expr","reduce_REFER","get_ref");
         /*  */
     install_keyword(DEREF,"deref");
     install_reduction(DEREF, assert_expr,get_val,reduce_DEREF );
     install_reduction_w(DEREF,"assert_expr","get_val","reduce_DEREF");
         /*  */
     install_keyword(LOC,"loc");
     install_reduction(LOC, assert_expr,reduce_const,get_ref );
     install_reduction_w(LOC,"assert_expr","reduce_const","get_ref");
         /*  */
     install_keyword(RUN,"run");
     install_reduction(RUN, reduce_RUN,assert_stmt,assert_stmt );
     install_reduction_w(RUN,"reduce_RUN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(EXIT,"exit");
     install_reduction(EXIT, reduce_EXIT,assert_stmt,assert_stmt );
     install_reduction_w(EXIT,"reduce_EXIT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(LOAD,"load");
     install_reduction(LOAD, reduce_LOAD,assert_stmt,assert_stmt );
     install_reduction_w(LOAD,"reduce_LOAD","assert_stmt","assert_stmt");
         /*  */
     install_keyword(DEFINE,"define");
     install_reduction(DEFINE, reduce_DEFINE,assert_stmt,assert_stmt );
     install_reduction_w(DEFINE,"reduce_DEFINE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SET,"set");
     install_reduction(SET, reduce_SET,assert_stmt,assert_stmt );
     install_reduction_w(SET,"reduce_SET","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SHOWTRACE,"showtrace");
     install_reduction(SHOWTRACE, reduce_SHOWTRACE,assert_stmt,assert_stmt );
     install_reduction_w(SHOWTRACE,"reduce_SHOWTRACE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SHOWMEM,"showmem");
     install_reduction(SHOWMEM, reduce_SHOWMEM,assert_stmt,assert_stmt );
     install_reduction_w(SHOWMEM,"reduce_SHOWMEM","assert_stmt","assert_stmt");
         /*  */
     install_keyword(ABORT,"abort");
     install_reduction(ABORT, reduce_ABORT,assert_stmt,assert_stmt );
     install_reduction_w(ABORT,"reduce_ABORT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(BREAK,"break");
     install_reduction(BREAK, reduce_BREAK,assert_stmt,assert_stmt );
     install_reduction_w(BREAK,"reduce_BREAK","assert_stmt","assert_stmt");
         /*  */
     install_keyword(RETURN,"return");
     install_reduction(RETURN, reduce_RETURN,assert_stmt,assert_stmt);
     install_reduction_w(RETURN,"reduce_RETURN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(CONTINUE,"continue");
     install_reduction(CONTINUE, reduce_CONTINUE,assert_stmt,assert_stmt );
     install_reduction_w(CONTINUE,"reduce_CONTINUE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(LEAP,"leap");
     install_reduction(LEAP, reduce_LEAP,assert_stmt,assert_stmt );
     install_reduction_w(LEAP,"reduce_LEAP","assert_stmt","assert_stmt");
         /*  */
     install_keyword(STEP,"step");
     install_reduction(STEP, reduce_STEP,assert_stmt,assert_stmt );
     install_reduction_w(STEP,"reduce_STEP","assert_stmt","assert_stmt");
         /*  */
     install_keyword(PEEK,"peek");
     install_reduction(PEEK, reduce_PEEK,assert_stmt,assert_stmt );
     install_reduction_w(PEEK,"reduce_PEEK","assert_stmt","assert_stmt");
         /*  */
     install_keyword(FORMAT,"format");
     install_reduction(FORMAT, reduce_FORMAT,assert_stmt,assert_stmt );
     install_reduction_w(FORMAT,"reduce_FORMAT","assert_stmt","assert_stmt");
         /*  */
     install_keyword(ASCII,"ascii");
     install_reduction(ASCII, assert_expr,reduce_ASCII,get_ref );
     install_reduction_w(ASCII,"assert_expr","reduce_ASCII","get_ref");
         /*  */
     install_keyword(HELP,"help");
     install_reduction(HELP, reduce_HELP,assert_stmt,assert_stmt );
     install_reduction_w(HELP,"reduce_HELP","assert_stmt","assert_stmt");
         /*  */
     install_keyword(SYSTEM,"system");
     install_reduction(SYSTEM, reduce_SYSTEM,assert_stmt,assert_stmt );
     install_reduction_w(SYSTEM,"reduce_SYSTEM","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iSYSTEM,"_system");
     install_reduction(iSYSTEM, reduce_iSYSTEM,assert_stmt,assert_stmt );
     install_reduction_w(iSYSTEM,"reduce_iSYSTEM","assert_stmt","assert_stmt");
         /*  */
     install_keyword(PASSLIMIT,"pass_limit");
     install_reduction(PASSLIMIT, assert_expr,reduce_SYSVAR,get_ref);
     install_reduction_w(PASSLIMIT,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(VERBOSITY,"verbosity");
     install_reduction(VERBOSITY, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(VERBOSITY,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(PRINTLEVEL,"print_level");
     install_reduction(PRINTLEVEL, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(PRINTLEVEL,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(DEBUGLEVEL,"debug_level");
     install_reduction(DEBUGLEVEL, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(DEBUGLEVEL,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(STATEPRINT,"print_states");
     install_reduction(STATEPRINT, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(STATEPRINT,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(REDUNDANCY,"no_reduncancy");
     install_reduction(REDUNDANCY, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(REDUNDANCY,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(CALLBYNAME,"call_by_name");
     install_reduction(CALLBYNAME, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(CALLBYNAME,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(ASSIGNAHEAD,"assign_ahead");
     install_reduction(ASSIGNAHEAD, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(ASSIGNAHEAD,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(BREAKISABORT,"break_is_abort");
     install_reduction(BREAKISABORT, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(BREAKISABORT,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(PATH,"path");
     install_reduction(PATH, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(PATH,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(HELPDIR,"help_dir");
     install_reduction(HELPDIR, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(HELPDIR,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(TIMESTEP,"time_step");
     install_reduction(TIMESTEP, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(TIMESTEP,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(STATENUMBER,"State_number");
     install_reduction(STATENUMBER, assert_expr,reduce_STATENUMBER,get_ref );
     install_reduction_w(STATENUMBER,"assert_expr","reduce_STATENUMBER","get_ref");
         /*  */
     install_keyword(RANDOM,"random");
     install_reduction(RANDOM, assert_expr,reduce_RANDOM,get_ref );
     install_reduction_w(RANDOM,"assert_expr","reduce_RANDOM","get_ref");
         /*  */
     install_keyword(RANDSEED,"random_seed");
     install_reduction(RANDSEED, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(RANDSEED,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(RRANDOM,"Random");
     install_reduction(RRANDOM, assert_expr,reduce_RRANDOM,get_ref );
     install_reduction_w(RRANDOM,"assert_expr","reduce_RRANDOM","get_ref");
         /*  */
         /*install_keyword(RRANDSEED,"Random_seed");
     install_reduction(RRANDSEED, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(RRANDSEED,"assert_expr","reduce_SYSVAR","get_ref");*/
         /*  */
     install_keyword(MAXRANDLEN,"max_randlen");
     install_reduction(MAXRANDLEN, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(MAXRANDLEN,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(IMPLIES,"implies");
     install_reduction(IMPLIES, reduce_IMPLIES,reduce_IMPLIES,reduce_IMPLIES );
     install_reduction_w(IMPLIES,"reduce_IMPLIES","reduce_IMPLIES","reduce_IMPLIES");
         /*  */
     install_keyword(MORE,"more");
     install_reduction(MORE, reduce_MORE_stmt,reduce_MORE_val,get_ref );
     install_reduction_w(MORE,"reduce_MORE","reduce_MORE_val","get_ref");
         /*  */
     install_keyword(CHOPSTAR,"chopstar");
     install_reduction(CHOPSTAR, reduce_CHOPSTAR,assert_stmt,assert_stmt );
     install_reduction_w(CHOPSTAR,"reduce_CHOPSTAR","assert_stmt","assert_stmt");
         /*  */
     install_keyword(iIF,"_if");
     install_reduction(iIF, assert_expr,reduce_iIF_val,get_ref );
     install_reduction_w(iIF,"assert_expr","reduce_iIF_val","get_ref");
         /*  */
     install_keyword(iUNDEFINED,"_undefined");
     install_reduction(iUNDEFINED, assert_expr,reduce_iUNDEFINED,get_ref );
     install_reduction_w(iUNDEFINED,"assert_expr","reduce_iUNDEFINED","get_ref");
         /*  */
     install_keyword(SHOWSTATE,"showstate");
     install_reduction(SHOWSTATE, reduce_SHOWSTATE,assert_stmt,assert_stmt );
     install_reduction_w(SHOWSTATE,"reduce_SHOWSTATE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(INFILE,"infile");
     install_reduction(INFILE, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(INFILE,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(OUTFILE,"outfile");
     install_reduction(OUTFILE, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(OUTFILE,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(RANLEN,"randlen");
     install_reduction(RANLEN, reduce_RANLEN,assert_stmt,assert_stmt );
     install_reduction_w(RANLEN,"reduce_RANLEN","assert_stmt","assert_stmt");
         /*  */
     install_keyword(INFINITE,"infinite");
     install_reduction(INFINITE, reduce_INFINITE,assert_stmt,assert_stmt );
     install_reduction_w(INFINITE,"reduce_INFINITE","assert_stmt","assert_stmt");
         /*  */
     install_keyword(HOPCHOP,"hopchop");
     install_reduction(HOPCHOP, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(HOPCHOP,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(PRECISION,"precision");
     install_reduction(PRECISION, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(PRECISION,"assert_expr","reduce_SYSVAR","get_ref");
         /*  */
     install_keyword(CEIL,"ceil");
     install_reduction(CEIL, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(CEIL,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(FLOOR,"floor");
     install_reduction(FLOOR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(FLOOR,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(RINT,"rint");
     install_reduction(RINT, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(RINT,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(SQRT,"sqrt");
     install_reduction(SQRT, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(SQRT,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ITOF,"itof");
     install_reduction(ITOF, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ITOF,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(EXPO,"exp");
     install_reduction(EXPO, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(EXPO,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(LOG,"log");
     install_reduction(LOG, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(LOG,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(LOGTEN,"log10");
     install_reduction(LOGTEN, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(LOGTEN,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(SIN,"sin");
     install_reduction(SIN, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(SIN,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(COS,"cos");
     install_reduction(COS, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(COS,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(TAN,"tan");
     install_reduction(TAN, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(TAN,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ASIN,"asin");
     install_reduction(ASIN, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ASIN,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ACOS,"acos");
     install_reduction(ACOS, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ACOS,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ATAN,"atan");
     install_reduction(ATAN, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ATAN,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ATANTWO,"atan2");
     install_reduction(ATANTWO, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(ATANTWO,"assert_expr","reduce_bin_op","get_ref");
         /*  */
     install_keyword(SINH,"sinh");
     install_reduction(SINH, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(SINH,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(COSH,"cosh");
     install_reduction(COSH, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(COSH,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(TANH,"tanh");
     install_reduction(TANH, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(TANH,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(FABS,"fabs");
     install_reduction(FABS, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(FABS,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(SRANDOM,"frandom");
     install_reduction(SRANDOM, assert_expr,reduce_SRANDOM,get_ref );
     install_reduction_w(SRANDOM,"assert_expr","reduce_SRANDOM","get_ref");
         /*  */
         /*install_keyword(SRANDSEED,"frandom_seed");
     install_reduction(SRANDSEED, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(SRANDSEED,"assert_expr","reduce_SYSVAR","get_ref");*/
         /*  */
     install_keyword(SRRANDOM,"fRandom");
     install_reduction(SRRANDOM, assert_expr,reduce_SRRANDOM,get_ref );
     install_reduction_w(SRRANDOM,"assert_expr","reduce_SRRANDOM","get_ref");
         /*  */
         /*install_keyword(SRRANDSEED,"fRandom_seed");
     install_reduction(SRRANDSEED, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(SRRANDSEED,"assert_expr","reduce_SYSVAR","get_ref");*/
         /*  */
     install_keyword(iOR,"_or");
     install_reduction(iOR, reduce_iOR,assert_stmt,assert_stmt );
     install_reduction_w(iOR,"reduce_iOR","assert_stmt","assert_stmt");
         /*  */
     install_keyword(DEFAULTS,"defaults");
     install_reduction(DEFAULTS, reduce_DEFAULTS,assert_stmt,assert_stmt );
     install_reduction_w(DEFAULTS,"reduce_DEFAULTS","assert_stmt","assert_stmt");
         /*  */
     install_keyword(MEM,"mem");
     install_reduction(MEM, reduce_MEM,assert_stmt,assert_stmt );
     install_reduction_w(MEM,"reduce_MEM","assert_stmt","assert_stmt");
         /*  */
     install_keyword(PREV,"prev");
     install_reduction(PREV, reduce_PREV_stmt,reduce_PREV_val,reduce_PREV_ref );
     install_reduction_w(PREV,"reduce_PREV_stmt","reduce_PREV_val","reduce_PREV_ref");
         /*  */
     install_keyword(NPREV,"nprev");
     install_reduction(NPREV, reduce_NPREV_stmt,reduce_NPREV_val,reduce_NPREV_ref );
     install_reduction_w(NPREV,"reduce_NPREV_stmt","reduce_NPREV_val","reduce_NPREV_ref");
         /*  */
     install_keyword(WINPUT,"winput");
     install_reduction(WINPUT, reduce_WINPUT,assert_stmt,assert_stmt );
     install_reduction_w(WINPUT,"reduce_WINPUT","assert_stmt","assert_stmt");
        /*  */
     install_keyword(iREMEMBER,"_remember");
     install_reduction(iREMEMBER, reduce_iREMEMBER,assert_stmt,assert_stmt );
     install_reduction_w(iREMEMBER,"reduce_iREMEMBER","assert_stmt","assert_stmt");
        /*  */
     install_keyword(MPZ,"big_integer");
     install_reduction(MPZ, assert_expr,reduce_const,get_ref );
     install_reduction_w(MPZ,"assert_expr","reduce_const","get_ref");
        /*  */
     install_keyword(MPX,"big_hexadecimal");
     install_reduction(MPX, assert_expr,reduce_const,get_ref );
     install_reduction_w(MPX,"assert_expr","reduce_const","get_ref");
        /*  */
     install_keyword(MPZTOI,"mpztoi");
     install_reduction(MPZTOI, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPZTOI,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPZTOMPX,"mpztompx");
     install_reduction(MPZTOMPX, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPZTOMPX,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPXTOI,"mpxtoi");
     install_reduction(MPXTOI, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPXTOI,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPXTOMPZ,"mpxtompz");
     install_reduction(MPXTOMPZ, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPXTOMPZ,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ITOMPZ,"itompz");
     install_reduction(ITOMPZ, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ITOMPZ,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(ITOMPX,"itompx");
     install_reduction(ITOMPX, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ITOMPX,"assert_expr","reduce_u_op","get_ref");
         /*  */
     install_keyword(MPFR,"big_float");
     install_reduction(MPFR, assert_expr,reduce_const,get_ref );
     install_reduction_w(MPFR,"assert_expr","reduce_const","get_ref");
        /*  */
     install_keyword(MPFRPRECISION,"mpfr_precision");
     install_reduction(MPFRPRECISION, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(MPFRPRECISION,"assert_expr","reduce_SYSVAR","get_ref");
        /*  */
     install_keyword(ITOMPFR,"itompfr");
     install_reduction(ITOMPFR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ITOMPFR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPZTOMPFR,"mpztompfr");
     install_reduction(MPZTOMPFR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPZTOMPFR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(ITOSTR,"itostr");
     install_reduction(ITOSTR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(ITOSTR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOVAL,"strtoval");
     install_reduction(STRTOVAL, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOVAL,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOI,"strtoi");
     install_reduction(STRTOI, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOI,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOF,"strtofl");
     install_reduction(STRTOF, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOF,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOMPZ,"strtompz");
     install_reduction(STRTOMPZ, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOMPZ,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOMPX,"strtompx");
     install_reduction(STRTOMPX, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOMPX,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(STRTOMPFR,"strtompfr");
     install_reduction(STRTOMPFR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(STRTOMPFR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(FTOSTR,"fltostr");
     install_reduction(FTOSTR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(FTOSTR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPZTOSTR,"mpztostr");
     install_reduction(MPZTOSTR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPZTOSTR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPXTOSTR,"mpxtostr");
     install_reduction(MPXTOSTR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPXTOSTR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPFRTOSTR,"mpfrtostr");
     install_reduction(MPFRTOSTR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPFRTOSTR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPFRTOF,"mpfrtofl");
     install_reduction(MPFRTOF, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPFRTOF,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(FTOMPFR,"fltompfr");
     install_reduction(FTOMPFR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(FTOMPFR,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPXTOMPFR,"mpxtompfr");
     install_reduction(MPXTOMPFR, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPXTOMPFR,"assert_expr","reduce_u_op","get_ref");
       /*  */
     install_keyword(MPXAND,"hex_and");
     install_reduction(MPXAND, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MPXAND,"assert_expr","reduce_bin_op","get_ref");
        /*  */
     install_keyword(MPXIOR,"hex_ior");
     install_reduction(MPXIOR, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MPXIOR,"assert_expr","reduce_bin_op","get_ref");
        /*  */
     install_keyword(MPXXOR,"hex_xor");
     install_reduction(MPXXOR, assert_expr,reduce_bin_op,get_ref );
     install_reduction_w(MPXXOR,"assert_expr","reduce_bin_op","get_ref");
        /*  */
     install_keyword(MPXCOM,"hex_com");
     install_reduction(MPXCOM, assert_expr,reduce_u_op,get_ref );
     install_reduction_w(MPXCOM,"assert_expr","reduce_u_op","get_ref");
        /*  */
     install_keyword(MPFRRND,"mpfr_rnd");
     install_reduction(MPFRRND, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(MPFRRND,"assert_expr","reduce_SYSVAR","get_ref");
        /*  */
     install_keyword(FLRND,"fl_rnd");
     install_reduction(FLRND, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(FLRND,"assert_expr","reduce_SYSVAR","get_ref");
        /*  */
     install_keyword(ZRANDSEED,"mp_random_seed");
     install_reduction(ZRANDSEED, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(ZRANDSEED,"assert_expr","reduce_SYSVAR","get_ref");
        /*  */
     install_keyword(ZRANDOMMAX,"zrandom_max");
     install_reduction(ZRANDOMMAX, assert_expr,reduce_SYSVAR,get_ref );
     install_reduction_w(ZRANDOMMAX,"assert_expr","reduce_SYSVAR","get_ref");
        /*  */
     install_keyword(ZRANDOM,"zrandom");
     install_reduction(ZRANDOM, assert_expr,reduce_ZRANDOM,get_ref );
     install_reduction_w(ZRANDOM,"assert_expr","reduce_ZRANDOM","get_ref");
        /*  */
     install_keyword(ZRRANDOM,"zRandom");
     install_reduction(ZRRANDOM, assert_expr,reduce_ZRRANDOM,get_ref );
     install_reduction_w(ZRRANDOM,"assert_expr","reduce_ZRRANDOM","get_ref");
        /*  */
     install_keyword(FRANDOM,"frrandom");
     install_reduction(FRANDOM, assert_expr,reduce_FRANDOM,get_ref );
     install_reduction_w(FRANDOM,"assert_expr","reduce_FRANDOM","get_ref");
        /*  */
     install_keyword(FRRANDOM,"frRandom");
     install_reduction(FRRANDOM, assert_expr,reduce_FRRANDOM,get_ref );
     install_reduction_w(FRRANDOM,"assert_expr","reduce_FRRANDOM","get_ref");
       /*  */
     install_keyword(COMMA,",");
     install_reduction(COMMA, noop, noop, noop );
     install_reduction_w(COMMA,"noop","noop","noop");
         /*  */
     install_keyword(TRUE,"true");
         /*  */
     install_keyword(FALSE,"false");
         /*  */
     install_keyword(NAME,"_name");
         /*  */
     install_keyword(THEN,"then");
         /*  */
     install_keyword(ELSE,"else");
         /*  */
     install_keyword(IN,"in");
         /*  */
     install_keyword(TIMES,"times");
         /*  */
     install_keyword(ENV,"_env");
         /*  */
     install_keyword(BND,"_bnd");
         /*  */
     install_keyword(TO,"to");
         /*  */
     install_keyword(DO,":");
         /*  */
     install_keyword(LPAREN,"(");
         /*  */
     install_keyword(RPAREN,")");
         /*  */
     install_keyword(LBRACE,"{");
         /*  */
     install_keyword(RBRACE,"}");
         /*  */
     install_keyword(LBRACK,"[");
         /*  */
     install_keyword(RBRACK,"]");
         /*  */
     install_keyword(BAR,"|");
         /*  */
     install_keyword(UNTIL,"until");
         /*  */
     install_keyword(HISTORY,"#");
         /*  */
     install_keyword(MEMBLOCK,"_memory");
         /*  */
     install_keyword(ELSEIF,"elseif");
         /*install_keyword(DO,strsave("do"));
     install_keyword(ALWAYS,strsave("box"));
     install_keyword(SOMETIME,strsave("sometime"));
     install_keyword(LEAP,strsave("jump"));
     install_keyword(NEXT,strsave("shift"));
     install_keyword(TO,strsave(".."));*/
}


void init_random (void)
{
  
  gmp_randinit_default (Randstate);
  mpz_init(ZRandSeed);
  mpz_set_str(ZRandSeed, "0", 10);
  /*gmp_randseed (Randstate, ZRandSeed);*/
  mpz_init(ZRandomMax);
  mpz_set_str(ZRandomMax, "999999999999999999", 10);
  mpz_init(ZRandomVal);
  mpz_init(ZRRandomVal);
  mpz_urandomm (ZRandomVal, Randstate, ZRandomMax);
  mpz_urandomm (ZRRandomVal, Randstate, ZRandomMax);
  mpfr_init(FRandomVal);
  mpfr_init(FRRandomVal);
  mpfr_urandom (FRandomVal, Randstate, MPFRRnd);
  mpfr_urandom (FRRandomVal, Randstate, MPFRRnd);
}

/* initialize tempura */
void tempura_init (void)
{
    catch_signals ();
        /*install_keywords ();*/
    initialise_symboltable();
    init_sysvars ();
    init_random();
    new_loc2 (TempLoc, STATIC, 1, 0);
    put_mem1 (new_loc2 (EmptyLoc, STATIC, 1, 0), True);
}


