/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: io.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"


/******************************\
*                              *
*  Auxiliary output routines.  *
*                              *
\******************************/

char *titoa(int64_t i)
{
   static char buf[MAXINTSIZE];
   char *p = buf;
   snprintf(buf,MAXINTSIZE,"%" PRId64 "",i);
   return p;
 }

void output_int(int64_t n, int f)
{
    char    s[MAXINTSIZE],
           *p = s;
    snprintf(s,MAXINTSIZE,"%" PRId64 "",n);
    outputf_str (p, f);
}

void foutput_int(int64_t n, int f)
{
    char    s[MAXINTSIZE],
            *p=s;
    
    snprintf(s,MAXINTSIZE,"%" PRId64 "",n);
    foutputf_str (p, f);
}


void output_str(char *s)
{
    int c;
    
    while (*s != '\0') {
        if (DebugLevel>2) {
            switch (c=*s++) {
                case '\n' :
                    output_char('\\');
                    output_char('n');
                    break;
                case '\r' :
                    output_char('\\');
                    output_char('r');
                    break;
                case '\t' :
                    output_char('\\');
                    output_char('t');
                    break;
                case '\b' :
                    output_char('\\');
                    output_char('b');
                    break;
                case '\f' :
                    output_char('\\');
                    output_char('f');
                    break;
                case '\\' :
                    output_char('\\');
                    output_char('\\');
                    break;
                case '"' :
                    output_char('\\');
                    output_char('"');
                    break;
                default:
                    output_char (c);
                    break;
            }
        }
        else
            output_char (*s++);
    }
    
}

void foutput_str(char *s)
{
    int c;
    while (*s != '\0') {
        if (DebugLevel>2) {
            switch (c=*s++) {
                case '\n' :
                    foutput_char('\\');
                    foutput_char('n');
                    break;
                case '\r' :
                    foutput_char('\\');
                    foutput_char('r');
                    break;
                case '\t' :
                    foutput_char('\\');
                    foutput_char('t');
                    break;
                case '\b' :
                    foutput_char('\\');
                    foutput_char('b');
                    break;
                case '\f' :
                    foutput_char('\\');
                    foutput_char('f');
                    break;
                case '\\' :
                    foutput_char('\\');
                    foutput_char('\\');
                    break;
                case '"' :
                    foutput_char('\\');
                    foutput_char('"');
                    break;    
                default:
                    foutput_char (c);
                    break;
            }
        }
        else
            foutput_char (*s++);
    }
}

void output_fl(double n)
{
  /*char    s[MAXFLSIZE];*/
  
  char *s2,*fmt,*p,*out1,*out2;
  char* out=NULL;
  int ret;
    
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    fmt = strapnd("%.",p);
    
    ret = asprintf(&out,fmt,n);
   
   if (ret < 0) {
     error_break0("cannot convert float to string");
   }
   
    out1 = strapnd("$",out);
    out2 = strapnd(out1,"$");
    free_mem(fmt);
    free_mem(p);
    free_mem(s2);
    
    output_str(out2);

    free_mem(out);
    free_mem(out1);
    free_mem(out2);
}

void output_mpz_abs(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 10)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 10, n);
  
  if (t[0] == '-') {
    t1 = strapnd("-0z",substr(t,1,strlen(t)-1));
    output_str(t1);
  } else {
    t1 = strapnd("0z",t);
    output_str(t1);
  }
  free_mem(t1);
}

void output_mpx_abs(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 16)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 16, n);
  
  if (t[0] == '-') {
    t1 = strapnd("0x",substr(t,1,strlen(t)-1));
    output_str(t1);
  } else {
    t1 = strapnd("0x", t);
    output_str(t1);
  }
  free_mem(t1);
} 

void output_mpz(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 10)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 10, n);

  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0z",substr(t,1,strlen(t)-1));
    output_str(t1);
  } else {
    t1 = strapnd("0z",t);
    output_str(t1);
  }
  
  free_mem(t1);
  
}

void output_mpx(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 16)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 16, n);

  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0x",substr(t,1,strlen(t)-1));
    output_str(t);
  } else {
    t1 = strapnd("0x",t);
    output_str(t1);
  }
  free_mem(t1);
  
}  

void output_mpfr(mpfr_t n)
{
  /*size_t size;
  size = max(mpfr_get_str_ndigits (10, MPFRPrecision)+2,7)+20;
  char t[size];

  char *s2,*fmt,*p;

  printf("begin output_mpfr\n");
  
  printf ("\n%zu\n",MPFRPrecision);

  printf ("\n%zu\n",mpfr_get_str_ndigits (10, MPFRPrecision));

  printf("output_mpfr: size = %zu\n", size);
  
  s2 = strsave(titoa(MPFRPrecision));
  p = strapnd(s2,"Re");
  fmt = strapnd("%.",p);
  */
  /*mpfr_sprintf(t, fmt, n);*/

  char *s2,*fmt,*p,*out1,*out2;
  char* out= NULL;
  int ret;

  s2 = strsave(titoa(MPFRPrecision));
  p = strapnd(s2,"Re");
  fmt = strapnd("%.",p);

  ret = mpfr_asprintf(&out,fmt,n);

  if (ret < 0) {
   error_break0("cannot convert big float to string");
  }
  
  out1 = strapnd("$$",out);
  out2 = strapnd(out1,"$$");
  output_str(out2);
  
  mpfr_free_str(out);
  free_mem(fmt);
  free_mem(p);
  free_mem(s2);
  free_mem(out1);
  free_mem(out2);
  /*mpfr_printf("%Re",n);*/
    
  /*output_str(t);*/

  /*free_mem(fmt);
  free_mem(p);
  free_mem(s2);*/

  /*printf("end output_mpfr\n");*/
  
}

void foutput_mpfr(mpfr_t n)
{
  /*size_t size;


    printf("begin foutput_mpfr\n");*/
  /*printf ("\n%zu\n",MPFRPrecision);

    printf ("\n%zu\n",mpfr_get_str_ndigits (10, MPFRPrecision));*/
  
  /*size = max(mpfr_get_str_ndigits (10, MPFRPrecision)+2,7)+20;*/

  /*printf("foutput_mpfr: size = %zu\n", size);*/
  
  /*char t[size];*/

  char *s2,*fmt,*p,*out1,*out2;
  char* out= NULL;
  int ret;

  s2 = strsave(titoa(MPFRPrecision));
  p = strapnd(s2,"Re");
  fmt = strapnd("%.",p);

  /*mpfr_sprintf(t, fmt, n);*/

  /*mpfr_get_str(t, e, 10, 0, n, MPFR_RNDN);*/
  /* use mpfr_sprintf */
  
  /*foutput_str(t);*/

  ret = mpfr_asprintf(&out,fmt,n);

  if (ret < 0) {
   error_break0("cannot convert big float to string");
  }
  
  out1 = strapnd("$$",out);
  out2 = strapnd(out1,"$$");
  foutput_str(out2);
  mpfr_free_str(out);
  free_mem(out1);
  free_mem(out2);
  
  free_mem(fmt);
  free_mem(p);
  free_mem(s2);

  /*printf("end output_mpfr\n");*/
  
}  


void foutput_mpz(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 10)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 10, n);
  
  
  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0z",substr(t,1,strlen(t)-1));
    foutput_str(t1);
  } else {
    t1 = strapnd("0z",t);
    foutput_str(t1);
  }

  free_mem(t1);
  
}

void foutput_mpx(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 16)+2;
  
  char t[size];
  char *t1;

  mpz_get_str(t, 16, n);
  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0x",substr(t,1,strlen(t)-1));
    foutput_str(t1);
  } else {
    t1 = strapnd("0x",t);
    foutput_str(t1);
  }

  free_mem(t1);
}


void foutput_mpz_abs(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 10)+2;

  /*printf("foutput_mpz: size = %zu\n", size);*/
  
  char t[size];

  mpz_get_str(t, 10, n);
  /*gmp_printf("\nfoutput_mpz:n %Zd\n", n);
    printf("foutput_mpz: t == %s\n", t);*/

  if (t[0] == '-') {
    foutput_str(substr(t,1,strlen(t)-1));
  } else {
    foutput_str(t);
  }
  
}

void foutput_mpx_abs(mpz_t n)
{
  size_t size;

  size = mpz_sizeinbase(n, 16)+2;

  /*printf("foutput_mpz: size = %zu\n", size);*/
  
  char t[size];

  mpz_get_str(t, 16, n);
  /*gmp_printf("\nfoutput_mpz:n %Zd\n", n);
    printf("foutput_mpz: t == %s\n", t);*/

  if (t[0] == '-') {
    foutput_str(substr(t,1,strlen(t)-1));
  } else {
    foutput_str(t);
  }
  
} 

void foutput_fl(double n)
{
  /*char    s[MAXFLSIZE];*/
  char *s2,*fmt,*p,*out1,*out2;
  char* out=NULL;
  int ret;
    
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    fmt = strapnd("%.",p);
    
    ret = asprintf(&out,fmt,n);

    if (ret < 0) {
      error_break0("cannot convert float to string");
    }
    
    out1 = strapnd("$",out);
    out2 = strapnd(out1,"$");
    
    free_mem(fmt);
    free_mem(p);
    free_mem(s2);
    
    foutput_str(out2);

    free_mem(out);
    free_mem(out1);
    free_mem(out2);
}

void outputf_str(char *s, int f)
{
    int     len = strlen (s);
    spaces (f - len);
    output_str (s);
    spaces (-f - len);
}

void foutputf_str(char *s, int f)
{
    int     len = strlen (s);
    fspaces (f - len);
    foutput_str (s);
    fspaces (-f - len);
}

void outputf_fl(double n, int f)
{
  /*char    s[MAXFLSIZE];*/
  char *s1,*s2,*fmt,*p,*p1,*p2,*out1,*out2;
  char* out=NULL;
  int ret;
    
    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    
    ret = asprintf(&out,fmt,n);

    if (ret < 0) {
      error_break0("cannot convert float to string");
    }
    
    out1 = strapnd("$",out);
    out2 = strapnd(out1,"$");
    
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    
    outputf_str(out2,f);
    
    free_mem(out);
    free_mem(out1);
    free_mem(out2);
}

void foutputf_fl(double n, int f)
{
  /*char    s[MAXFLSIZE];*/
  char *s1,*s2,*fmt,*p,*p1,*p2,*out1,*out2;
  char* out=NULL;
  int ret;
    
    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    
    ret = asprintf(&out,fmt,n);

    if (ret < 0) {
      error_break0("cannot convert float to string");
    }
    
    out1 = strapnd("$",out);
    out2 = strapnd(out1,"$");
    
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    
    foutputf_str(out2,f);
    

    free_mem(out);
    free_mem(out1);
    free_mem(out2);
}

void foutputf_fl_f(double n, int f)
{
  /*char    s[MAXFLSIZE];*/
  char *s1,*s2,*fmt,*p,*p1,*p2,*out1,*out2;

  char* out = NULL;
  int ret;
    
    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"f");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    
    ret = asprintf(&out,fmt,n);

    if (ret < 0) {
      error_break0("cannot convert float to string");
    }
    
    out1 = strapnd("$",out);
    out2 = strapnd(out1,"$");
    
    
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    
    foutputf_str(out2,f);

    free_mem(out);
    free_mem(out1);
    free_mem(out2);
}

void foutputf_mpz_f(mpz_t n, int f)
{
  size_t size;

  size = mpz_sizeinbase(n, 10)+2;

  /*printf("foutput_mpz: size = %zu\n", size);*/
  
  char t[size];
  char *t1;

  mpz_get_str(t, 10, n);
  
  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0z",substr(t,1,strlen(t)-1));
  } else {
    t1 = strapnd("0z",t);
  }

  foutputf_str(t1,f);
  
  
  /*gmp_printf("\nfoutput_mpz:n %Zd\n", n);
    printf("foutput_mpz: t == %s\n", t);*/
  free_mem(t1);

  
} 


void foutputf_mpx_f(mpz_t n, int f)
{
  size_t size;

  size = mpz_sizeinbase(n, 16)+2;

  /*printf("foutput_mpz: size = %zu\n", size);*/
  
  char t[size];
  char *t1;

  mpz_get_str(t, 16, n);
  
  if (mpz_sgn(n) == -1) {
    t1 = strapnd("-0x",substr(t,1,strlen(t)-1));
  } else {
    t1 = strapnd("0x",t);
  }

  foutputf_str(t1, f);
  
  /*gmp_printf("\nfoutput_mpz:n %Zd\n", n);
    printf("foutput_mpz: t == %s\n", t);*/
  free_mem(t1);

  
} 


void foutputf_mpfr_f(mpfr_t n, int f)
{
  /*size_t size;
  int written;

  printf("begin foutput_mpfr_f\n");*/
  
  /*printf ("\n%zu\n",MPFRPrecision);

    printf ("\n%zu\n",mpfr_get_str_ndigits (10, MPFRPrecision));*/

  /*printf (" exp = %ld %ld\n", mpfr_get_exp(n), mpfr_get_emax());*/
  
  /*size = max(mpfr_get_str_ndigits (10, MPFRPrecision)+2,7)+20;*/

  /*printf("foutput_mpfr: size = %zu\n", size);*/
  
  /*char t[size];*/

  
  char *s1,*s2,*fmt,*p,*p1,*p2,*out1,*out2;
  char* out = NULL;
  int ret;
    
  s1 = strsave(titoa(f));
  s2 = strsave(titoa(MPFRPrecision));
  p = strapnd(s2,"Re");
  p1 = strapnd(".",p);
  p2 = strapnd(s1,p1);
  fmt = strapnd("%",p2);

  /*printf("fmt = %s\n", fmt);*/

   ret = mpfr_asprintf(&out, fmt, n);

   if (ret < 0) {
     error_break0("cannot convert big float to string");
   }
   
   out1 = strapnd("$$",out);
   out2 = strapnd(out1,"$$");

  /*printf (" t = %s \n",t);*/
  
  /*printf( "written = %d\n",written);*/
  
  /*mpfr_get_str(t, e, 10, 0, n, MPFR_RNDN);*/
  /* use mpfr_sprintf */

  foutputf_str(out2,f);

  mpfr_free_str(out);
  
  
  free_mem(fmt);
  free_mem(p);
  free_mem(p1);
  free_mem(p2);
  free_mem(s1);
  free_mem(s2);
  free_mem(out1);
  free_mem(out2);

  /*printf("end foutputf_mpfr_f\n");*/
  
}  


void output_lambda(NODEPTR e)
{
    output_expn(e,-1);
}

void foutput_lambda(NODEPTR e)
{
    foutput_expn(e,-1);
}

/* Find print-name of type. */
char *print_name(NODEPTR e)
{
    unsigned t = node_type(e) == 0 ? token_type(END) : token_type(node_type(e));
    static char *undef = "undefined";
    return ((t <= token_type(MAXTOKEN)) ? Keyword[t] : undef);
}

void message(char *text)
{
    printf ("***Tempura: %s.\n", text);
    fflush(stdout);
}

void spaces(int n)
{
    while (n-- > 0)
        output_char (' ');
}

void fspaces(int n)
{
    while (n-- > 0)
        foutput_char (' ');
}

void output_env(NODEPTR env)
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        output_str (node_nm_name(id));
        printf(" (%d,%d)",node_nm_level(id),node_nm_call_level_offset(id));
        if ((id = node_next(id)) != NULL)
            output_str (", ");
    }
}

void foutput_env(NODEPTR env)
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        foutput_str (node_nm_name(id));
        /*foutput_int (node_nm_level(id),-1);*/
        if ((id = node_next(id)) != NULL)
            foutput_str (", ");
    }
}

void output_expn(NODEPTR e, int level)
{
  NODEPTR bnd_1;
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int lambda_level, k;
  
    if (e == NULL) {
/*        output_str ("()");*/
        output_str("?");
        return;
    }

    if (level == 0) {
        output_str ("...");
        return;
    }

    switch (node_type(e)) {
        
        case NAME:
            output_char ('(');
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_str (node_nm_name(e));
            output_char (')');
            break;
        case STATE:
        case STATIC:
            output_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:  
        case MPFR:  
        case LIST:
        case LOC:
            output_val (e);
            break;
        case BND:
            bnd_1 = e;
            output_str("BND: ");
            bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
            for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                output_str (node_nm_name(id));
                output_str (" = ");
                loc = bnd_blk_base_loc + node_nm_off(id);
                if (is_indirect_mem(loc)) {
                  dereference(loc);
                }
                if (1) {
                  output_expn (node_next(curr_loc(loc)), -1);
                } else {
                  output_str("( Curr: ");
                  output_expn (node_next(curr_loc(loc)), -1);
                  output_str(", Next: ");
                  output_expn (node_next(next_loc(loc)), -1);
                  output_str(", Prev: ");
                  for(k=2;k<=1+node_mem_history(loc);k = k+1) {
                    output_str("(");
                    output_int(k-1,0);
                    output_str(") ");
                    output_expn (node_next(prev_loc1(loc,k)), -1);
                    if (k < 1+node_mem_history(loc)) {
                      output_str("; ");
                    }         
                  }
                  output_str(") ");
                }
                if (node_next(id) != NULL)
                  output_str (", ");
              }
            while (node_bnd_next(bnd_1) != NULL) {
              output_str(" ; ");
              bnd_1 = node_bnd_next(bnd_1);
              bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
            for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                output_str (node_nm_name(id));
                output_str (" = ");
                loc = bnd_blk_base_loc + node_nm_off(id);
                if (1) {
                  dereference(loc);
                }
                if (1) {
                  output_expn (node_next(curr_loc(loc)), -1);
                } else {
                  output_str("( Curr: ");
                  output_expn (node_next(curr_loc(loc)), -1);
                  output_str(", Next: ");
                  output_expn (node_next(next_loc(loc)), -1);
                  output_str(", Prev: ");
                  for(k=2;k<=1+node_mem_history(loc);k = k+1) {
                    output_str("(");
                    output_int(k-1,0);
                    output_str(") ");
                    output_expn (node_next(prev_loc1(loc,k)), -1);
                    if (k < 1+node_mem_history(loc)) {
                      output_str("; ");
                    }         
                  }
                  output_str(") ");
                }
                if (node_next(id) != NULL)
                  output_str (", ");
              }
            }
          /*output_expn (node_bnd_env(e), level);*/
            /*while (node_bnd_next(bnd_1) != NULL) {
              output_char(',');
              output_expn(node_bnd_env(bnd_1), level);
              bnd_1 = node_bnd_next(bnd_1);
              }   */           
            break;
        case ENV:
            output_env (e); 
            break;
        case MEMBLOCK:
            output_str (Keyword[token_type(node_type(e))]);
            output_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:    
        case MORE:    
            output_str (Keyword[token_type(node_type(e))]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:  
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            output_char ('(');
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case MPXAND:
        case MPXIOR:
        case MPXXOR:     
        case ATANTWO:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level);
            output_char (',');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;    
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
        case SKIPUASSN:    
        case DEFAULTS:
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            break;
        case iUNDEFINED:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level);
            output_char (')');
            break;
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case iLEN:    
        case STB:
        case MEM:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case iREMEMBER:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_1(e), level);
            break;
        case NEXT:
        case WNEXT:
        case LEN:    
        case PREV:    
        case CEIL:
        case FLOOR:
        case RINT:  
        case SQRT:
        case ITOF:
        case MPZTOMPFR:
        case MPXTOMPFR:  
        case ITOMPZ:
        case ITOMPX:  
        case MPZTOI:
        case MPZTOMPX:  
        case MPXTOI:
        case MPXTOMPZ:
        case ITOSTR:
        case STRTOVAL:  
        case STRTOI:
        case STRTOF:
        case STRTOMPZ:
        case STRTOMPX:
        case STRTOMPFR:  
        case FTOSTR:
        case MPZTOSTR:
        case MPXTOSTR:
        case MPFRTOSTR:
        case MPFRTOF:
        case FTOMPFR:  
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case MPXCOM:   
        case FORMAT:
            output_str (Keyword[token_type(node_type(e))]);
	    output_char ('(');;
            output_expn (node_lnk_1(e), level);
	    output_char (')');;
            break;
        case IF:
        case iIF:    
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" then (");
            output_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                output_str (") else (");
                output_expn (node_lnk_3(e), level - 1);
            }
            output_char (')');
            break;  
        case CALL:
        case iCALL:
            output_expn (node_lnk_1(e), level);
            output_char ('(');
            output_expn (node_lnk_2(e), level);
            output_char (')');
            break;
        case iLAMBDA:
          /*
            output_char ('(');
            output_expn (node_lnk_1(e), level);
            output_str (" with ");
            output_expn (node_lnk_1(e), level);
            output_char (')');
          */
            output_val(e);
            break;
        case LAMBDA:
            output_char ('(');
            output_str (Keyword[token_type(node_type(e))]);
            output_char (' ');
            output_expn (node_lnk_1(e), level);
            output_str (" : ");
            output_expn (node_lnk_2(e), level - 1);
            /*output_str (" )");*/
            output_char (')');
            /*lambda_level = (node_env_idlist(node_lnk_1(e)) != NULL) ?
              node_nm_level(node_env_idlist(node_lnk_1(e))) : 0;*/
              
            /*printf("lambda_level = %d\n", lambda_level);*/

            /*if (node_lnk_3(e) != NULL) {
              output_str (" with ");
              bnd_1 = node_lnk_3(e);
              bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));*/
              /*output_expn (node_bnd_env(bnd_1), level );*/
              /*for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                output_str (node_nm_name(id));
                output_str (" = ");
                if (mem_defined1 ( (bnd_blk_base_loc + node_nm_off(id))))
                  output_val(get_mem((bnd_blk_base_loc + node_nm_off(id))));
                else foutput_char('?');
                if (node_next(id) != NULL)
                  output_str (", ");
              }
              while (node_bnd_next(bnd_1) != NULL) {
                bnd_1 = node_bnd_next(bnd_1);
                bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
                if (node_env_idlist(node_bnd_env(bnd_1)) != NULL)
                  if (node_nm_level(node_env_idlist(node_bnd_env(bnd_1))) < lambda_level) {
                  output_char(';'); */
                    /*output_expn (node_bnd_env(bnd_1), level );*/
            /*
                    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                         id = node_next(id)) {
                     output_str (node_nm_name(id));
                     output_str (" = ");
                     if (mem_defined1 ( (bnd_blk_base_loc + node_nm_off(id))))
                       output_val(get_mem((bnd_blk_base_loc + node_nm_off(id))));
                     else output_char('?');
                     if (node_next(id) != NULL)
                       output_str (", ");
                    }
                }
              }
            }
            */
            /*
            if (node_lnk_3(e) != NULL) {
              output_str (" with ");
              output_env (node_bnd_env(e));
              }
            */  
            break;
        case EXISTS:
        case FORALL:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char ('<');
                output_expn (node_lnk_2(e), level);
            }
            output_str (": (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case WHILE:
            output_str (Keyword[token_type(node_type(e))]);
            output_char (' ');
            output_expn (node_lnk_1(e), level);
            output_str (" do (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case FOR:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('<');
            output_expn (node_lnk_2(e), level);
            output_str (" do (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case FORIN:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" in ");
            output_expn (node_lnk_2(e), level);
            output_str (" do (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case FORALLIN:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" in ");
            output_expn (node_lnk_2(e), level);
            output_str (" : (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case FORTIMES:
            output_str ("for ");
            output_expn (node_lnk_1(e), level);
            output_str (" times do (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case DEFINE:
        case SET:
            output_str ("(");
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('=');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case COMMA:
	    output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char (',');
		output_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" : (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case iAND:
        case iOR:
        case iSEMI:
        case iPROJ:
            output_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                space;
                output_str (Keyword[token_type(node_type(e))]);
                space;
                output_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            output_char ('[');
            output_expn (node_lnk_1(e), level - 1);
            output_char (']');
            break;
        case MAKERANGELIST:
            output_char ('[');
            output_expn (node_lnk_1(e), level - 1);
            output_str (" : ");
            output_expn (node_lnk_2(e), level);
            output_char ('<');
            output_expn (node_lnk_3(e), level);
            output_char (']');
            break;
        case SUB:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_char (']');
            break;
        case SIZE:
            output_char ('|');
            output_expn (node_lnk_1(e), level);
            output_char ('|');
            break;
        case SUBLIST:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_str ("..");
            output_expn (node_lnk_3(e), level - 1);
            output_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case iSYSTEM:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level - 1);
            output_char (',');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;    
        default:
            output_str (print_name (e));
            break;
    }
}

void foutput_expn(NODEPTR e, int level)
{
  NODEPTR bnd_1;
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int lambda_level, k;
  
    if (e == NULL) {
/*        output_str ("()");*/
        foutput_str ("?");
        return;
    }

    if (level == 0) {
        foutput_str ("...");
        return;
    }

    switch (node_type(e)) {

        case STATE:
        case STATIC:
            foutput_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:  
        case MPFR:
        case LIST:
        case LOC:
            foutput_val (e);
            break;
        case BND:
            bnd_1 = e;
            foutput_str("BND: ");
            bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
            for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                foutput_str (node_nm_name(id));
                foutput_str (" = ");
                loc = bnd_blk_base_loc + node_nm_off(id);
                if (is_indirect_mem(loc)) {
                  dereference(loc);
                }
                if (1) {
                  foutput_expn (node_next(curr_loc(loc)), -1);
                } else {
                  foutput_str("( Curr: ");
                  foutput_expn (node_next(curr_loc(loc)), -1);
                  foutput_str(", Next: ");
                  foutput_expn (node_next(next_loc(loc)), -1);
                  foutput_str(", Prev: ");
                  for(k=2;k<=1+node_mem_history(loc);k = k+1) {
                    foutput_str("(");
                    foutput_int(k-1,0);
                    foutput_str(") ");
                    foutput_expn (node_next(prev_loc1(loc,k)), -1);
                    if (k < 1+node_mem_history(loc)) {
                      foutput_str("; ");
                    }         
                  }
                  foutput_str(") ");
                }
                if (node_next(id) != NULL)
                  foutput_str (", ");
              }
            while (node_bnd_next(bnd_1) != NULL) {
              foutput_str(" ; ");
              bnd_1 = node_bnd_next(bnd_1);
              bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
            for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                foutput_str (node_nm_name(id));
                foutput_str (" = ");
                loc = bnd_blk_base_loc + node_nm_off(id);
                if (is_indirect_mem(loc)) {
                  dereference(loc);
                }
                if (1) {
                  foutput_expn (node_next(curr_loc(loc)), -1);
                } else {
                  foutput_str("( Curr: ");
                  foutput_expn (node_next(curr_loc(loc)), -1);
                  foutput_str(", Next: ");
                  foutput_expn (node_next(next_loc(loc)), -1);
                  foutput_str(", Prev: ");
                  for(k=2;k<=1+node_mem_history(loc);k = k+1) {
                    foutput_str("(");
                    foutput_int(k-1,0);
                    foutput_str(") ");
                    foutput_expn (node_next(prev_loc1(loc,k)), -1);
                    if (k < 1+node_mem_history(loc)) {
                      foutput_str("; ");
                    }         
                  }
                  foutput_str(") ");
                }
                if (node_next(id) != NULL)
                  foutput_str (", ");
              }
            }
            /*
            while (node_bnd_next(bnd_1) != NULL) {
                bnd_1 = node_bnd_next(bnd_1);
                bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
                if (node_env_idlist(node_bnd_env(bnd_1)) != NULL) {
                 
                    foutput_char(';'); 
                    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                         id = node_next(id)) {
                     foutput_str (node_nm_name(id));
                     foutput_str (" = ");
                     if (mem_defined1 ( (bnd_blk_base_loc + node_nm_off(id))))
                       foutput_val(get_mem((bnd_blk_base_loc + node_nm_off(id))));
                     else foutput_char('?');
                     if (node_next(id) != NULL)
                       foutput_str (", ");
                    }
                }
              }
            */
          /*foutput_expn (node_bnd_env(e), level);*/
            /*while (node_bnd_next(bnd_1) != NULL) {
              foutput_char(',');
              bnd_1 = node_bnd_next(bnd_1);
              foutput_expn(node_bnd_env(bnd_1), level);
              } */
            break;
        case ENV:
            foutput_env (e); 
            break;
        case MEMBLOCK:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:
        case MORE:    
            foutput_str (Keyword[token_type(node_type(e))]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER: 
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case MPXAND:
        case MPXIOR:
        case MPXXOR:      
        case ATANTWO:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
            foutput_char (',');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
        case SKIPUASSN:    
        case DEFAULTS:
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            break;
        case iUNDEFINED:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
            foutput_char (')');
            break;    
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case iLEN:
        case STB:
        case MEM:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case iREMEMBER:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_1(e), level);
            break;
        case NEXT:
        case WNEXT:
        case LEN:    
        case PREV:    
        case CEIL:
        case RINT:  
        case FLOOR:
        case SQRT:
        case ITOF:
        case MPZTOI:
        case MPZTOMPX:  
        case MPXTOI:
        case MPXTOMPZ:  
        case ITOMPZ:
        case ITOMPX:  
        case MPZTOMPFR:
        case MPXTOMPFR:
        case ITOSTR:
        case STRTOVAL:  
        case STRTOI:
        case STRTOF:
        case STRTOMPZ:
        case STRTOMPX:
        case STRTOMPFR:  
        case FTOSTR:
        case MPZTOSTR:
        case MPXTOSTR:
        case MPFRTOSTR:
        case MPFRTOF:
        case FTOMPFR:  
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case MPXCOM:   
        case FORMAT:
            foutput_str (Keyword[token_type(node_type(e))]);
	    foutput_char ('(');;
            foutput_expn (node_lnk_1(e), level);
	    foutput_char (')');;
            break;
        case IF:
        case iIF:    
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" then (");
            foutput_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                foutput_str (") else (");
                foutput_expn (node_lnk_3(e), level - 1);
            }
            foutput_char (')');
            break;   
        case CALL:
        case iCALL:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('(');
            foutput_expn (node_lnk_2(e), level);
            foutput_char (')');
            break;
        case iLAMBDA:
          /*foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" with ");
            foutput_expn (node_lnk_1(e), level);
            foutput_char (')');*/
            foutput_val(e);
            break;    
        case LAMBDA:
            foutput_char ('(');
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char (' ');
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" : ");
            foutput_expn (node_lnk_2(e), level - 1);
            /*foutput_str (" )");*/
            foutput_char (')');
            /*
            lambda_level = (node_env_idlist(node_lnk_1(e)) != NULL) ?
            node_nm_level(node_env_idlist(node_lnk_1(e))) : 0;*/
              
            /*printf("lambda_level = %d\n", lambda_level);*/

            /*
            if (node_lnk_3(e) != NULL) {
              foutput_str (" with ");
              bnd_1 = node_lnk_3(e);
              bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
            */
              /*foutput_expn (node_bnd_env(bnd_1), level );*/
            /*
              for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
                foutput_str (node_nm_name(id));
                foutput_str (" = ");
                if (mem_defined1 ( (bnd_blk_base_loc + node_nm_off(id))))
                  foutput_val(get_mem((bnd_blk_base_loc + node_nm_off(id))));
                else foutput_char('?');
                if (node_next(id) != NULL)
                  foutput_str (", ");
              }
              while (node_bnd_next(bnd_1) != NULL) {
                bnd_1 = node_bnd_next(bnd_1);
                bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
                if (node_env_idlist(node_bnd_env(bnd_1)) != NULL)
                  if (node_nm_level(node_env_idlist(node_bnd_env(bnd_1))) < lambda_level) {
                  foutput_char(';'); */
                    /*foutput_expn (node_bnd_env(bnd_1), level );*/
            /*
                    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                         id = node_next(id)) {
                     foutput_str (node_nm_name(id));
                     foutput_str (" = ");
                     if (mem_defined1 ( (bnd_blk_base_loc + node_nm_off(id))))
                       foutput_val(get_mem((bnd_blk_base_loc + node_nm_off(id))));
                     else foutput_char('?');
                     if (node_next(id) != NULL)
                       foutput_str (", ");
                    }
                }
              }
            }
            */
            /*
            if (node_lnk_3(e) != NULL) {
              foutput_str (" with ");
              bnd_1 = node_lnk_3(e);
              foutput_env (node_bnd_env(bnd_1));
              while (node_bnd_next(bnd_1) != NULL) {
                foutput_char(';');
                
                bnd_1 = node_bnd_next(bnd_1);
                foutput_env (node_bnd_env(bnd_1));
              }
            }
            */
            break;
        case EXISTS:
        case FORALL:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char ('<');
                foutput_expn (node_lnk_2(e), level);
            }
            foutput_str (": (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case WHILE:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char (' ');
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case FOR:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('<');
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case FORIN:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" in ");
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case FORALLIN:
            foutput_str (Keyword[token_type(node_type(e))]);
            space;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" in ");
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" : (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;   
        case FORTIMES:
            foutput_str ("for ");
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" times do (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;    
        case DEFINE:
        case SET:
            output_char ('(');
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('=');
            foutput_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case COMMA:
	    foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char (',');
		foutput_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" : (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case iAND:
        case iOR:
        case iSEMI:
        case iPROJ:
            foutput_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                fspace;
                foutput_str (Keyword[token_type(node_type(e))]);
                fspace;
                foutput_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            foutput_char ('[');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (']');
            break;
        case MAKERANGELIST:
            foutput_char ('[');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_str (" : ");
            foutput_expn (node_lnk_2(e), level);
            foutput_char ('<');
            foutput_expn (node_lnk_3(e), level);
            foutput_char (']');
            break;    
        case SUB:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (']');
            break;
        case SIZE:
            foutput_char ('|');
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('|');
            break;
        case SUBLIST:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_str ("..");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case iSYSTEM:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (',');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;    
        default:
            foutput_str (print_name (e));
            break;
    }
}

void foutput_char (int c)
{
    char   name[MAXFILENAMESIZE];
    FILE * fp = NULL;
    if (strcmp (node_strval(OutFile), "stdout") == 0) {
        output_char (c);
    }
    else {
        if ((fp = openfile (name, node_strval(Path), node_strval(OutFile), FXOUT, "a")) == NULL) {
            warn1 ("cannot open file \"%s\"",  node_strval(OutFile));
        warn0 ("will display to stdout");
        output_char (c);
        }
        else {
            putc(c,fp);
           CharsOutput++;
           fclose (fp);
        }
    }
}
