/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: system.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"


#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#else /* HAVE_SYS_WAIT_H */
#if !defined(WNOHANG) /* && (defined(bsd43) || defined(vax)) */
#define WNOHANG 1
#endif
#ifndef WEXITSTATUS
#define WEXITSTATUS(X) ((unsigned)(X) >> 8)
#endif
#ifndef WIFEXITED
#define WIFEXITED(X) (((X) & 255) == 0)
#endif
#ifndef WIFSIGNALED
#define WIFSIGNALED(X) ((((X) & 255) != 0x255 && ((X) & 255) != 0))
#endif
#ifndef WTERMSIG
#define WTERMSIG(X) ((X) & 255)
#endif
#endif /* HAVE_SYS_WAIT_H */

/* For profiling on SUN */
/* char *strrchr(a,b)
char *a,b;
{return (char *) rindex(a,b);}
*/
/****************************\
*                            *
*  System specific routines  *
*                            *
\****************************/


/******************\
*                  *
*  Error handling  *
*                  *
\******************/
void onintr(int sig)
{
    signal (SIGINT, onintr);
    InterruptSet = T;
/*    user_break ("interrupt");*/
}

void onfpe(int sig)
{
    signal (SIGFPE, onfpe);
    error_break0 ("division by 0 exception");
}

/* SIGBUS for djgpp */
#ifdef SIGBUS
void onbus(int sig)
{
    signal (SIGBUS, onbus);
    error_break0 ("bus error");
}
#endif

void onsegv(int sig)
{
    signal (SIGSEGV, onsegv);
    error_break0 ("segmentation violation");
    
}
/*  SIGSYS (for bad system calls) not available on Linux?  BCM, 4Dec96 */
#ifdef SIGSYS
void onsys(int sig)
{
    signal (SIGSYS, onsys);
    error_break0 ("bad argument to system call");
}
#endif

#ifdef SIGALRM
void onalrm(int sig)
{
    signal (SIGALRM, onalrm);
    AlarmGone = T;
}
#endif

void catch_signals(void)
{
    if (signal (SIGINT, SIG_IGN) != SIG_IGN)
        signal (SIGINT, onintr);
    if (signal (SIGFPE, SIG_IGN) != SIG_IGN)
        signal (SIGFPE, onfpe);
/* SIGBUS for djgpp */
#ifdef SIGBUS
    if (signal (SIGBUS, SIG_IGN) != SIG_IGN)
        signal (SIGBUS, onbus);
#endif    
    if (signal (SIGSEGV, SIG_IGN) != SIG_IGN)
        signal (SIGSEGV, onsegv);
/*  SIGSYS (for bad system calls) not available on Linux?  BCM, 4Dec96 */
#ifdef SIGSYS    
    if (signal (SIGSYS, SIG_IGN) != SIG_IGN)
        signal (SIGSYS, onsys);
#endif
#ifdef SIGALRM    
    if (signal (SIGALRM, SIG_IGN) != SIG_IGN)
        signal (SIGALRM, onalrm);
#endif
/*    intrinit (onintr, INTERRUPTSTACK, CNTRL_C); FOR PC */
}

void breakin(void)
{
    int c,val;
    InterruptSet = F;
    if (BreakIsAbortFlag) {
        fputs ("   Abort\n", stdout);
        fflush(stdout);
        TempuraLevel = 0;
        restore_state (BreakPoint,val);
    }
    else {    
    fputs ("   Abort (a), Break (b) or Continue (c)?\n ", stdout);
    fflush(stdout);
    while (isspace (c = getchar ()));
    while (T) {
        while (getchar () != '\n');
                                /* Discard up to newline. */
        switch (c) {
            case 'a':
            case 'A':
                putchar ('\n');
                TempuraLevel = 0;
                restore_state (BreakPoint,val);
            case 'b':
            case 'B':
                putchar ('\n');
                tempura (++TempuraLevel);
                return;
            case 'c':
            case 'C':
                putchar ('\n');
                return;
            default:
                fputs ("   Please type a, b or c: ", stdout);
                fflush(stdout);
        }
        c = getchar ();
    }
    }
}



/******************\
*                  *
*   File handling  *
*                  *
\******************/
/* Modified by 05/09/2001 to handle absolute paths better */
FILE *openfile(char *fullname, char *spath, char *name, char *suffix, char *mode)
{
    char   *p ;
    FILE   *fp;
    free (allocate (1000, 1));  /* check enough space for buffer */
    if ( name[0]==FILENAMESEP ) {  /* check for absolute path.  BCM 28Jul97 */
      strcpy(fullname,name);
      strcat(fullname,suffix);
      fp = fopen (fullname, mode);
    }
    else {
        do {
            *(p = fullname) = '\0';
            while (*spath != '\0' && !fisspace ((int)*spath)) {
                *p++ = *spath++;
                *p = FILENAMESEP;
            }
            *++p = '\0';
            while (fisspace ((int)*spath))
                spath++;
            strcat (fullname, name);
            strcat (fullname, suffix);
        } while (((fp = fopen (fullname, mode)) == NULL) && (*spath != '\0'));
    };
    return (fp);
}

int fisspace (int s)
{
    if (s == ';') return(1);
    else return(0);
}




/*******************************\
*                               *
*  Initialise system variables  *
*                               *
\*******************************/

void fl_set_rnd (int val)
{
  switch (val) {
  case 0:
    fesetround(FE_TONEAREST);
    break;
  case 1:
    fesetround(FE_DOWNWARD);
    break;
  case 2:
    fesetround(FE_UPWARD);
    break;
  case 3:
    fesetround(FE_TOWARDZERO);
    break;
  default:
    fesetround(FE_TONEAREST);
    break;
  }
}

void mpfr_set_rnd (int val)
{
  switch (val) {
  case 0 :
    MPFRRnd = MPFR_RNDN;
    break;
  case 1:
    MPFRRnd = MPFR_RNDD;
    break;
  case 2:
    MPFRRnd = MPFR_RNDU;
    break;
  case 3:
    MPFRRnd = MPFR_RNDZ;
    break;
  case 4:
    MPFRRnd = MPFR_RNDA;
    break;
  case 5:
    MPFRRnd = MPFR_RNDF;
    break;
  default:
    MPFRRnd = MPFR_RNDN;
    break;
  }
  mpfr_set_default_rounding_mode (MPFRRnd);
}

void assign_sysvar(NODEPTR id, NODEPTR val)
{
    switch (node_type(id)) {
	case PASSLIMIT:
            PassLimit = node_ival(val);
            break;
	case VERBOSITY:
            Verbosity = node_ival(val);
            break;
	case PRINTLEVEL:
            PrintLevel = node_ival(val);
            break;
	case DEBUGLEVEL:
            DebugLevel = node_ival(val);
            break;
	case STATEPRINT:
            StatePrintFlag = node_bval(val);
            break;
	case REDUNDANCY:
            RedundancyFlag = node_bval(val);
            break;
	case CALLBYNAME:
            CallByNameFlag = node_bval(val);
            break;
	case ASSIGNAHEAD:
            AssignAheadFlag = node_bval(val);
            break;
        case BREAKISABORT:
            BreakIsAbortFlag = node_bval(val);
            break;
        case HOPCHOP:
            HopChop = node_bval(val);
            break;    
	case PATH:
            /* Next line changed to avoid garbarge collection problems
	       BCM 14Dec96 */
            Path = val;
                /* Path = stringp (val) -> str;*/
            break;
	case HELPDIR:
            /* Next line changed to avoid garbarge collection problems
	       BCM 14Dec96 */
            HelpDir = val;
                /* HelpDir = stringp (val) -> str;*/
	    break;
        case INFILE:
            OldInFile  = InFile;
            InFile = val;
            break;
        case OUTFILE:
            OutFile = val;
            break;    
	case TIMESTEP:
            TimeStep = node_ival(val);
            break;
        case RANDSEED:
            RandSeed = node_ival(val);
	    TRandSeed=(uint64_t)RandSeed;  
            init_xorshift128plus(TRandSeed);
	    RandomVal = -1;       /* Reset random.    */
	    RandomValTime = 0;
            RRandomVal = -1;
            RRandomValTime = 0;
            SRandomVal = -1;
            SRandomValTime = 0;
            SRRandomVal = -1;
            SRRandomValTime = 0; 
            break;
        case ZRANDOMMAX:
            mpz_set(ZRandomMax, node_mpzval(val));
            ZRandomValTime = 0;
            ZRRandomValTime = 0;
            FRandomValTime = 0;
            FRRandomValTime = 0;
            mpz_urandomm (ZRandomVal, Randstate, ZRandomMax);
            mpz_urandomm (ZRRandomVal, Randstate, ZRandomMax);
            mpfr_urandom (FRandomVal, Randstate, MPFRRnd);
            mpfr_urandom (FRRandomVal, Randstate, MPFRRnd);
            break;
        case ZRANDSEED:
            mpz_set(ZRandSeed, node_mpzval(val));
            gmp_randseed (Randstate, ZRandSeed);
            ZRandomValTime = 0;
            ZRRandomValTime = 0;
            FRandomValTime = 0;
            FRRandomValTime = 0;
            mpz_urandomm (ZRandomVal, Randstate, ZRandomMax);
            mpz_urandomm (ZRRandomVal, Randstate, ZRandomMax);
            mpfr_urandom (FRandomVal, Randstate, MPFRRnd);
            mpfr_urandom (FRRandomVal, Randstate, MPFRRnd);
            break;
                /*case RRANDSEED:
            RRandSeed = node_ival(val);
	    TRRandSeed=RRandSeed;  
	    RRandomVal = -1;
	    RRandomValTime = 0;
            break;
         case SRANDSEED:      
            SRandSeed = node_ival(val);
	    TSRandSeed=SRandSeed;  
	    SRandomVal = -1;
	    SRandomValTime = 0;
            break;
        case SRRANDSEED:      
            SRRandSeed = node_ival(val);
	    TSRRandSeed=SRRandSeed;  
	    SRRandomVal = -1;
	    SRRandomValTime = 0;
            break;   */
	case MAXRANDLEN:    /* Added AC 29APR 1999  */
	    if ( node_ival(val) >= 0)
                MaxRandLen = node_ival(val);
	    else {
	      printf("\n***Maximum random length must be >= 0.\n");
              fflush(stdout);
            }            
            break;
         case PRECISION:    /* Added AC 29APR 1999  */
	    if ( node_ival(val) >= 0 )
                Precision = node_ival(val);
	    else {
	      printf("\n***Precision must be >= 0.\n");
              fflush(stdout);
            }            
            break;
         case MPFRPRECISION:    
           if ( node_ival(val) >= 0) {
                MPFRPrecision = node_ival(val);
                mpfr_set_default_prec (MPFRPrecision);
           }
	    else {
	      printf("\n***mpfr_precision must be >= 0.\n");
              fflush(stdout);
            }            
            break;
         case MPFRRND:
           if ( node_ival(val) >= 0 && node_ival(val) <= 5) {
                TMPFRRnd = node_ival(val);
                mpfr_set_rnd (TMPFRRnd);
           } else {
              printf("\n***mpfr_rnd must be >= 0 and <= 5.\n");
              fflush(stdout);
            }            
            break;
          case FLRND:
           if ( node_ival(val) >= 0 && node_ival(val) <= 3) {
                FLRnd = node_ival(val);
                fl_set_rnd (FLRnd);
           } else {
              printf("\n***fl_rnd must be >= 0 and <= 3.\n");
              fflush(stdout);
            }            
            break;  
        default:
            break;
    }
}

void init_sysvars(void)
{
    /* The following was changed to avoid garbage collection problems.
       BCM 14Dec96 */
    Path = 
      string_node(strsave (getenv ("TPATH") == NULL ? "" : getenv ("TPATH")));
    /* Old version:
    Path = strsave (getenv ("TPATH") == NULL ? "" : getenv ("TPATH"));
    */
    /* Following added by BCM, 4Dec96, 14Dec96 */
    HelpDir = 
      string_node(strsave (getenv ("THELPDIR") == NULL ? "Help"
		       : getenv ("THELPDIR")));
    InFile = string_node(strsave ("stdin"));
    OutFile = string_node(strsave ("stdout"));
    /* */
    mpfr_set_default_prec (MPFRPrecision);
    mpfr_set_rnd (TMPFRRnd);
    fl_set_rnd (FLRnd);
}

void start_time(void)
{
    AlarmGone = F;
#ifdef SIGALRM
    alarm ((unsigned)TimeStep);
#endif
}


void check_time(void)
{   
#ifdef SIGALRM
    if (!AlarmGone) 
        pause ();
    else
        warn0 ("falling behind (real) time");
#endif
}


int timeval_subtract(struct timeval *result, struct timeval *t2, struct timeval *t1)
{
    long int diff = 1000000 * (t2->tv_sec - t1->tv_sec) + (t2->tv_usec - t1->tv_usec);
        /*(t2->tv_usec + 1000000 * t2->tv_sec) - (t1->tv_usec + 1000000 * t1->tv_sec);*/
    result->tv_sec = diff / 1000000;
    result->tv_usec = diff % 1000000;

    return (diff<0);
}

void timeval_print(struct timeval *tv)
{
    char buffer[30];
    time_t curtime;

#if defined(__APPLE__) && defined(__MACH__)
      printf("%ld.%06d", tv->tv_sec, tv->tv_usec);
#else
      printf("%ld.%06ld", tv->tv_sec, tv->tv_usec);
#endif  
    curtime = tv->tv_sec;
    strftime(buffer, 30, "%m-%d-%Y  %T", localtime(&curtime));
#if defined(__APPLE__) && defined(__MACH__)
    printf(" = %s.%06d\n", buffer, tv->tv_usec);
#else    
    printf(" = %s.%06ld\n", buffer, tv->tv_usec);
#endif    
}


void check_sys_waits (void)
{
    int pid;
    /*AC: the wait problem
      WAIT_STATUS_TYPE s;  
      WAIT_STATUS_TYPE status = &s */
    WAIT_STATUS_TYPE status; 
    /*int *statusp;*/
        /* waitpid (-1,&status, WNOHANG) for those
           that haven't got wait3 (&status, WNOHANG, 0)*/
#ifdef HAVE_WORKING_FORK
/*#ifdef NO_WAIT3*/
while ((pid = waitpid (-1,&status,WNOHANG)) != -1 && pid != 0) {
/*
  #else
   while ((pid = wait3 (&status,WNOHANG, 0) ) != -1 && pid != 0) {
   #endif    */   
SysWaitList = node2 (iSYSTEM, int_node (pid), SysWaitList);
assert2d (WIFEXITED(status), 
	         "system call %d terminated with non-zero status: %d",
		 pid, WEXITSTATUS(status) ); 
    }
#endif
}
